from functools import partial
from pathlib import Path

import torch
import torchaudio
from torch.utils.data import Dataset

from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)
from everyvoice.dataloader import BaseDataModule
from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.dataset import (
    FastSpeech2DataModule,
)
from everyvoice.text.lookups import LookupTables
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils import (
    _flatten,
    check_dataset_size,
    filter_dataset_based_on_target_text_representation_level,
)
from everyvoice.utils.heavy import get_segments


class E2EDataset(Dataset):
    def __init__(self, dataset, config: EveryVoiceConfig, use_segments=True):
        self.dataset = dataset
        self.config = config
        self.use_segments = use_segments
        self.lookup = LookupTables(config)
        self.sep = "--"
        self.text_processor = TextProcessor(config.feature_prediction.text)
        self.preprocessed_dir = Path(
            self.config.feature_prediction.preprocessing.save_dir
        )
        self.output_sampling_rate = (
            self.config.vocoder.preprocessing.audio.output_sampling_rate
        )
        self.input_sampling_rate = (
            self.config.feature_prediction.preprocessing.audio.input_sampling_rate
        )
        self.sampling_rate_change = (
            self.output_sampling_rate // self.input_sampling_rate
        )
        self.output_hop_size = (
            self.config.vocoder.preprocessing.audio.fft_hop_size
            * self.sampling_rate_change
        )
        self.frame_segment_size = (
            self.config.vocoder.preprocessing.audio.vocoder_segment_size
            // self.output_hop_size
        )
        self.speaker2id = self.lookup.speaker2id
        self.lang2id = self.lookup.lang2id

    def _load_file(self, bn, spk, lang, dir, fn):
        return torch.load(
            self.preprocessed_dir / dir / self.sep.join([bn, spk, lang, fn])
        )

    def _load_audio(self, bn, spk, lang, dir, fn):
        audio, _ = torchaudio.load(
            str(self.preprocessed_dir / dir / self.sep.join([bn, spk, lang, fn]))
        )
        return audio

    def __getitem__(self, index):
        """
        Returns dict with keys: {
            "mel"
            "duration"
            "pfs"
            "text"
            "raw_text"
            "basename"
            "speaker"
            "speaker_id"
            "language"
            "language_id"
            "label"
            "energy"
            "pitch"
            "audio"
            "audio_mel"
        }
        """
        item = self.dataset[index]
        speaker = "default" if "speaker" not in item else item["speaker"]
        language = "default" if "language" not in item else item["language"]
        speaker_id = self.speaker2id[speaker]
        language_id = self.lang2id[language]
        basename = item["basename"]
        mel = self._load_file(
            basename,
            speaker,
            language,
            "spec",
            f"spec-{self.input_sampling_rate}-{self.config.feature_prediction.preprocessing.audio.spec_type}.pt",
        ).transpose(
            0, 1
        )  # [mel_bins, frames] -> [frames, mel_bins]
        if self.config.feature_prediction.model.learn_alignment:
            match self.config.feature_prediction.model.target_text_representation_level:
                case TargetTrainingTextRepresentationLevel.characters:
                    duration = self._load_file(
                        basename,
                        speaker,
                        language,
                        "attn",
                        f"{DatasetTextRepresentation.characters.value}-attn-prior.pt",
                    )
                case TargetTrainingTextRepresentationLevel.ipa_phones | TargetTrainingTextRepresentationLevel.phonological_features:
                    duration = self._load_file(
                        basename,
                        speaker,
                        language,
                        "attn",
                        f"{DatasetTextRepresentation.ipa_phones.value}-attn-prior.pt",
                    )
                case _:
                    raise NotImplementedError(
                        f"{self.config.feature_prediction.model.target_text_representation_level} have not yet been implemented."
                    )
        else:
            duration = self._load_file(
                basename, speaker, language, "duration", "duration.pt"
            )
        match self.config.feature_prediction.model.target_text_representation_level:
            case TargetTrainingTextRepresentationLevel.characters:
                text = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["character_tokens"]
                    )
                )
            case TargetTrainingTextRepresentationLevel.ipa_phones | TargetTrainingTextRepresentationLevel.phonological_features:
                text = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["phone_tokens"]
                    )
                )
            case _:
                raise NotImplementedError(
                    f"{self.config.feature_prediction.model.target_text_representation_level} have not yet been implemented."
                )
        if TargetTrainingTextRepresentationLevel.characters.value in item:
            raw_text = item[TargetTrainingTextRepresentationLevel.characters.value]
        else:
            raw_text = item.get(
                TargetTrainingTextRepresentationLevel.ipa_phones.value, "text"
            )
        pfs = None
        if (
            self.config.feature_prediction.model.target_text_representation_level
            == TargetTrainingTextRepresentationLevel.phonological_features
        ):
            pfs = self._load_file(basename, speaker, language, "pfs", "pfs.pt")

        energy = self._load_file(basename, speaker, language, "energy", "energy.pt")
        pitch = self._load_file(basename, speaker, language, "pitch", "pitch.pt")
        audio = self._load_audio(
            basename,
            speaker,
            language,
            "audio",
            f"audio-{self.output_sampling_rate}.wav",
        )
        if audio.dim() == 1:
            audio = audio.unsqueeze(
                0
            )  # audio should have channel dimension for vocoder
        audio_mel = self._load_file(
            basename,
            speaker,
            language,
            "spec",
            f"spec-{self.output_sampling_rate}-{self.config.vocoder.preprocessing.audio.spec_type}.pt",
        )  # [mel_bins, frames]
        if self.use_segments:
            audio, start_sample = get_segments(
                audio, self.config.vocoder.preprocessing.audio.vocoder_segment_size
            )
            audio_mel, start_frame = get_segments(
                audio_mel, self.frame_segment_size, start_sample // self.output_hop_size
            )
            input_start_frame = (
                start_frame // self.sampling_rate_change
            )  # we need to take the corresponding segment from the output of the feature prediction network, so this stores the index to start from
        return {
            "mel": mel,
            "duration": duration,
            "pfs": pfs,
            "text": text,
            "raw_text": raw_text,
            "basename": basename,
            "speaker": speaker,
            "speaker_id": speaker_id,
            "language": language,
            "language_id": language_id,
            # "label": item["label"],
            "energy": energy,
            "pitch": pitch,
            "audio": audio,
            "audio_mel": audio_mel,
            "segment_start_frame": input_start_frame if self.use_segments else None,
        }

    def __len__(self):
        return len(self.dataset)

    def get_labels(self):
        return [x["label"] for x in self.dataset]


class E2EDataModule(BaseDataModule):
    def __init__(self, config: EveryVoiceConfig):
        super().__init__(config=config)
        self.collate_fn = partial(
            FastSpeech2DataModule.collate_method,
            learn_alignment=config.feature_prediction.model.learn_alignment,
        )
        self.use_weighted_sampler = (
            config.feature_prediction.training.use_weighted_sampler
        )
        self.batch_size = (
            config.training.batch_size
        )  # TODO: should this be set somewhere else?
        self.load_dataset()
        self.dataset_length = len(self.train_dataset) + len(self.val_dataset)

    def load_dataset(self):
        self.train_dataset = self.config.training.filelist_loader(
            self.config.training.training_filelist
        )
        self.val_dataset = self.config.training.filelist_loader(
            self.config.training.validation_filelist
        )

    def prepare_data(self):
        (
            self.train_dataset,
            self.val_dataset,
        ) = filter_dataset_based_on_target_text_representation_level(
            self.config.feature_prediction.model.target_text_representation_level,
            self.train_dataset,
            self.val_dataset,
            self.batch_size,
        )
        train_samples = len(self.train_dataset)
        val_samples = len(self.val_dataset)
        check_dataset_size(self.batch_size, train_samples, "training")
        check_dataset_size(self.batch_size, val_samples, "validation")
        self.train_dataset = E2EDataset(self.train_dataset, self.config)
        self.val_dataset = E2EDataset(self.val_dataset, self.config)
        # save it to disk
        torch.save(self.train_dataset, self.train_path)
        torch.save(self.val_dataset, self.val_path)



import json
from pathlib import Path

import questionary
from email_validator import EmailNotValidError, validate_email
from rich import print
from rich.panel import Panel
from rich.style import Style

from everyvoice.config.preprocessing_config import Dataset, PreprocessingConfig
from everyvoice.config.shared_types import (
    BaseTrainingConfig,
    ContactInformation,
    LoggerConfig,
    init_context,
)
from everyvoice.config.text_config import Symbols, TextConfig
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.e2e.config import E2ETrainingConfig, EveryVoiceConfig
from everyvoice.model.feature_prediction.config import (
    FastSpeech2ModelConfig,
    FeaturePredictionConfig,
)
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.utils import generic_psv_filelist_reader, slugify, write_filelist
from everyvoice.wizard import (
    ALIGNER_CONFIG_FILENAME_PREFIX,
    PREPROCESSING_CONFIG_FILENAME_PREFIX,
    SPEC_TO_WAV_CONFIG_FILENAME_PREFIX,
    TEXT_CONFIG_FILENAME_PREFIX,
    TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX,
    TEXT_TO_WAV_CONFIG_FILENAME_PREFIX,
    Step,
    StepNames,
)
from everyvoice.wizard.dataset import TextProcessingStep, get_dataset_steps
from everyvoice.wizard.prompts import (
    CUSTOM_QUESTIONARY_STYLE,
    get_response_from_menu_prompt,
)
from everyvoice.wizard.utils import write_dict_to_config


class NameStep(Step):
    DEFAULT_NAME = StepNames.name_step

    def prompt(self):
        return input(
            "What would you like to call this project? This name should reflect the model you intend to train, e.g. 'my-sinhala-project' or 'english-french-model' or something similarly descriptive of your project: "
        )

    def validate(self, response):
        if len(response) == 0:
            print("Sorry, your project needs a name. ")
            return False
        sanitized_path = slugify(response)
        if not sanitized_path == response:
            print(
                f"Sorry, the project name '{response}' is not valid, since it will be used to create a folder and special characters are not permitted for folder names. Please re-type something like '{sanitized_path}' instead."
            )
            return False
        return True

    def effect(self):
        print(
            f"Great! Launching Configuration Wizard 🧙 for project named '{self.response}'."
        )


class ContactNameStep(Step):
    DEFAULT_NAME = StepNames.contact_name_step

    def prompt(self):
        return input("What is your full name? ")

    def validate(self, response):
        # Some languages don't use first and last names, so we can't necessarily check that response.split() > 1
        # It would be nice to have a better check here though.
        if len(response) < 3:
            print("Sorry, EveryVoice requires a name to help prevent misuse.")
            return False
        return True

    def effect(self):
        print(f"Great! Nice to meet you, '{self.response}'.")


class ContactEmailStep(Step):
    DEFAULT_NAME = StepNames.contact_email_step

    def prompt(self):
        return input("Please provide a contact email address for your models. ")

    def in_unit_testing(self):
        """Skip checking deliverability when in unit testing.

        Checking deliverability can be slow where there is not web connection, as
        is sometimes the case when running the unit tests, so skip it in that context.
        """
        import inspect

        return any(
            frame.filename.endswith("test_wizard.py") for frame in inspect.stack()
        )

    def validate(self, response):
        try:
            # Check that the email address is valid. Turn on check_deliverability
            # for first-time validations like on account creation pages (but not
            # login pages).
            validate_email(response, check_deliverability=not self.in_unit_testing())
        except EmailNotValidError as e:
            # The exception message is a human-readable explanation of why it's
            # not a valid (or deliverable) email address.
            print("EveryVoice requires a valid email address to prevent misuse.")
            print(str(e))
            return False
        return True

    def effect(self):
        emailinfo = validate_email(self.response, check_deliverability=False)
        email = emailinfo.normalized
        self.response = email
        print(
            f"Great! Your contact email '{self.response}' will be saved to your models."
        )


class OutputPathStep(Step):
    DEFAULT_NAME = StepNames.output_step

    def prompt(self):
        return questionary.path(
            "Where should the Configuration Wizard save your files?",
            default=".",
            style=CUSTOM_QUESTIONARY_STYLE,
            only_directories=True,
        ).unsafe_ask()

    def sanitize_input(self, response):
        response = super().sanitize_input(response)
        return response.strip()

    def validate(self, response) -> bool:
        path = Path(response)
        if path.is_file():
            print(f"Sorry, '{path}' is a file. Please select a directory.")
            return False
        assert self.state is not None, "OutputPathStep requires NameStep"
        output_path = path / self.state.get(StepNames.name_step, "DEFAULT_NAME")
        if output_path.exists():
            print(
                f"Sorry, '{output_path}' already exists. "
                "Please choose another output directory or start again and choose a different project name."
            )
            return False

        # We create the output directory in validate() instead of effect() so that
        # failure can be reported to the user and the question asked again if necessary.
        try:
            output_path.mkdir(parents=True, exist_ok=True)
            # we created it just to test permission, but don't leave it lying around in
            # case the wizard is interrupted or fails. We'll create it again when we save.
            output_path.rmdir()
        except OSError as e:
            print(
                f"Sorry, could not create '{output_path}': {e}. "
                "Please choose another output directory."
            )
            return False

        self.output_path = output_path
        return True

    def effect(self):
        print(
            f"The Configuration Wizard 🧙 will put your files here: '{self.output_path}'"
        )


class ConfigFormatStep(Step):
    DEFAULT_NAME = StepNames.config_format_step

    def prompt(self):
        return get_response_from_menu_prompt(
            "Which format would you like to output the configuration to?",
            ("yaml", "json"),
        )

    def validate(self, response):
        return response in ("yaml", "json")

    def effect(self):
        output_path = (
            Path(self.state[StepNames.output_step]) / self.state[StepNames.name_step]
        ).expanduser()
        # create_config_files
        config_dir = output_path / "config"
        config_dir.absolute().mkdir(exist_ok=True, parents=True)
        # preprocessed dir
        preprocessed_dir = output_path / "preprocessed"
        preprocessed_dir.absolute().mkdir(parents=True, exist_ok=True)
        # used in configs
        preprocessed_dir_relative_to_configs = Path("..") / "preprocessed"
        # log dir
        log_dir = output_path / "logs_and_checkpoints"
        log_dir.absolute().mkdir(parents=True, exist_ok=True)
        log_dir_relative_to_configs = Path("..") / "logs_and_checkpoints"
        datasets = []
        # Text Configuration
        symbols = {}
        multispeaker = False
        multilingual = False
        global_cleaners = (
            []
        )  # TODO: this should be fixed by https://github.com/roedoejet/EveryVoice/issues/359
        for dataset in [key for key in self.state.keys() if key.startswith("dataset_")]:
            dataset_state = self.state[dataset]
            # Add Cleaners
            # TODO: these should really be dataset-specific cleaners, not global cleaners
            # so this should be fixed by https://github.com/roedoejet/EveryVoice/issues/359
            if dataset_state.get(StepNames.text_processing_step):
                global_cleaners += [
                    TextProcessingStep().process_lookup[x]["fn"]
                    for x in dataset_state[StepNames.text_processing_step]
                ]
            # Gather Symbols for Text Configuration
            # rename keys based on dataset name:
            dataset_name = dataset_state[StepNames.dataset_name_step]
            dataset_symbols = {
                f"{dataset_name}_{k}": v
                for k, v in dataset_state[StepNames.symbol_set_step].items()
            }
            symbols.update(dataset_symbols)
            if dataset_state.get(StepNames.data_has_language_value_step, "no") == "yes":
                multilingual = True
            if dataset_state.get(StepNames.data_has_speaker_value_step, "no") == "yes":
                multispeaker = True
            # Dataset Configs
            wavs_dir = Path(dataset_state[StepNames.wavs_dir_step]).expanduser()
            if not wavs_dir.is_absolute():
                if not output_path.is_absolute():
                    for _ in config_dir.parts:
                        wavs_dir = Path("..") / wavs_dir
                else:
                    wavs_dir = Path.cwd() / wavs_dir
            new_filelist_path = (
                Path("..") / f"{dataset_name}-filelist.psv"
            ).expanduser()
            filelist_data = dataset_state["filelist_data"]
            for i, entry in enumerate(filelist_data):
                # Remove .wav if it was added to the basename
                if entry["basename"].endswith(".wav"):
                    entry["basename"] = entry["basename"].replace(".wav", "")
                # Remove unknown columns
                filelist_data[i] = {
                    k: v
                    for k, v in entry.items()
                    if k is not None and not k.startswith("unknown")
                }
            write_filelist(filelist_data, (config_dir / new_filelist_path).absolute())
            sox_effects = dataset_state["sox_effects"]
            filelist_loader = generic_psv_filelist_reader

            datasets.append(
                Dataset(
                    label=dataset,
                    data_dir=wavs_dir,
                    filelist=new_filelist_path,
                    filelist_loader=filelist_loader,
                    sox_effects=sox_effects,
                    permissions_obtained=True,  # If you get this far, you've answered the Dataset Permission Attestation step correctly
                )
            )

        text_config = TextConfig(symbols=Symbols(**symbols))
        text_config.cleaners += global_cleaners
        text_config_path = Path(f"{TEXT_CONFIG_FILENAME_PREFIX}.{self.response}")
        write_dict_to_config(
            json.loads(text_config.model_dump_json(exclude_none=False)),
            (config_dir / text_config_path).absolute(),
        )
        # Contact
        CONTACT_INFO = ContactInformation(
            contact_name=self.state[StepNames.contact_name_step],
            contact_email=self.state[StepNames.contact_email_step],
        )

        with init_context({"writing_config": config_dir.resolve()}):
            # Preprocessing Config
            preprocessed_training_filelist_path = (
                preprocessed_dir_relative_to_configs / "training_filelist.psv"
            )
            preprocessed_validation_filelist_path = (
                preprocessed_dir_relative_to_configs / "validation_filelist.psv"
            )
            preprocessing_config = PreprocessingConfig(
                dataset=self.state[StepNames.name_step],
                save_dir=preprocessed_dir_relative_to_configs,
                source_data=datasets,
            )
            preprocessing_config_path = Path(
                f"{PREPROCESSING_CONFIG_FILENAME_PREFIX}.{self.response}"
            )
            write_dict_to_config(
                json.loads(preprocessing_config.model_dump_json(exclude_none=False)),
                (config_dir / preprocessing_config_path).absolute(),
            )

            ## Create Aligner Config
            aligner_logger = LoggerConfig(
                name="AlignerExperiment", save_dir=log_dir_relative_to_configs
            ).model_dump()
            aligner_config = AlignerConfig(
                contact=CONTACT_INFO,
                # This isn't the actual AlignerTrainingConfig, but we can use it because we just
                # inherit the defaults if we pass a dict to the AlignerConfig.training field
                training=BaseTrainingConfig(
                    training_filelist=preprocessed_training_filelist_path,
                    validation_filelist=preprocessed_validation_filelist_path,
                    logger=aligner_logger,
                ).model_dump(),
            )
            aligner_config_path = Path(
                f"{ALIGNER_CONFIG_FILENAME_PREFIX}.{self.response}"
            )
            aligner_config_json = json.loads(
                aligner_config.model_dump_json(
                    exclude_none=False, exclude={"preprocessing": True, "text": True}
                )
            )
            aligner_config_json["path_to_preprocessing_config_file"] = str(
                preprocessing_config_path
            )
            aligner_config_json["path_to_text_config_file"] = str(text_config_path)
            write_dict_to_config(
                aligner_config_json,
                (config_dir / aligner_config_path).absolute(),
            )

            # Create Feature Prediction Config
            fp_logger = LoggerConfig(
                name="FeaturePredictionExperiment", save_dir=log_dir_relative_to_configs
            )
            fp_config = FeaturePredictionConfig(
                contact=CONTACT_INFO,
                model=FastSpeech2ModelConfig(
                    multilingual=multilingual,
                    multispeaker=multispeaker,
                ),
                training=BaseTrainingConfig(
                    training_filelist=preprocessed_training_filelist_path,
                    validation_filelist=preprocessed_validation_filelist_path,
                    logger=fp_logger,
                ).model_dump(),
            )
            fp_config_path = Path(
                f"{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.{self.response}"
            )
            fp_config_json = json.loads(
                fp_config.model_dump_json(
                    exclude_none=False, exclude={"preprocessing": True, "text": True}
                )
            )
            fp_config_json["path_to_preprocessing_config_file"] = str(
                preprocessing_config_path
            )
            fp_config_json["path_to_text_config_file"] = str(text_config_path)
            write_dict_to_config(
                fp_config_json,
                (config_dir / fp_config_path).absolute(),
            )

            # Create Vocoder Config
            vocoder_logger = LoggerConfig(
                name="VocoderExperiment", save_dir=log_dir_relative_to_configs
            )
            vocoder_config = VocoderConfig(
                contact=CONTACT_INFO,
                training=BaseTrainingConfig(
                    training_filelist=preprocessed_training_filelist_path,
                    validation_filelist=preprocessed_validation_filelist_path,
                    logger=vocoder_logger,
                ).model_dump(),
            )
            vocoder_config_path = Path(
                f"{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}.{self.response}"
            )
            vocoder_config_json = json.loads(
                vocoder_config.model_dump_json(
                    exclude_none=False, exclude={"preprocessing": True}
                )
            )
            vocoder_config_json["path_to_preprocessing_config_file"] = str(
                preprocessing_config_path
            )
            write_dict_to_config(
                vocoder_config_json,
                (config_dir / vocoder_config_path).absolute(),
            )

            # E2E Config
            e2e_logger = LoggerConfig(
                name="E2E-Experiment", save_dir=log_dir_relative_to_configs
            )
            e2e_config = EveryVoiceConfig(
                contact=CONTACT_INFO,
                aligner=aligner_config,
                feature_prediction=fp_config,
                vocoder=vocoder_config,
                training=E2ETrainingConfig(
                    training_filelist=preprocessed_training_filelist_path,
                    validation_filelist=preprocessed_validation_filelist_path,
                    logger=e2e_logger,
                ).model_dump(),
            )
            e2e_config_json = json.loads(
                e2e_config.model_dump_json(
                    exclude_none=False,
                    exclude={
                        "aligner": True,
                        "feature_prediction": True,
                        "vocoder": True,
                    },
                )
            )
            e2e_config_json["path_to_aligner_config_file"] = str(aligner_config_path)
            e2e_config_json["path_to_feature_prediction_config_file"] = str(
                fp_config_path
            )
            e2e_config_json["path_to_vocoder_config_file"] = str(vocoder_config_path)
            e2e_config_path = Path(
                f"{TEXT_TO_WAV_CONFIG_FILENAME_PREFIX}.{self.response}"
            )
            write_dict_to_config(
                e2e_config_json,
                (config_dir / e2e_config_path).absolute(),
            )

        print(
            Panel(
                f"You've finished configuring your dataset. Your files are located at {config_dir.absolute()}",
                title="Congratulations 🎉",
                subtitle="Next Steps Documentation: https://docs.everyvoice.ca/guides",
                border_style=Style(color="#0B4F19"),
            )
        )


class MoreDatasetsStep(Step):
    DEFAULT_NAME = StepNames.more_datasets_step

    def prompt(self):
        return get_response_from_menu_prompt(
            "Do you have more datasets to process?",
            ("no", "yes"),
        )

    def validate(self, response):
        return response in ("yes", "no")

    def effect(self):
        if self.response == "yes":
            new_dataset_index = (
                max(
                    [
                        int(key.split("_")[1])
                        for key in self.state.keys()
                        if key.startswith("dataset_")
                    ],
                    default=0,
                )
                + 1
            )

            self.tour.add_steps(
                get_dataset_steps(dataset_index=new_dataset_index)
                + [MoreDatasetsStep()],
                self,
            )
        elif len([key for key in self.state.keys() if key.startswith("dataset_")]) == 0:
            print("No dataset to save, exiting without saving any configuration.")
        else:
            self.tour.add_step(
                ConfigFormatStep(name=StepNames.config_format_step), self
            )



from enum import Enum
from pathlib import Path

import typer
from everyvoice.base_cli.interfaces import complete_path


class BenchmarkType(str, Enum):
    training = "training"
    inference = "inference"


def benchmark(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    benchmark_type: BenchmarkType = BenchmarkType.training,
    gpu: bool = True,
    warmup_reps: int = 10,
    repetitions: int = 300,
):
    import time
    from functools import partial

    import numpy as np
    import torch

    from ..config import FastSpeech2Config
    from ..dataset import FastSpeech2DataModule
    from ..model import FastSpeech2

    config = FastSpeech2Config.load_config_from_path(config_file)
    loader = FastSpeech2DataModule(config)
    loader.prepare_data()
    batch = loader.collate_method(
        [loader.train_dataset[i] for i in range(config.training.batch_size)]
    )
    model = FastSpeech2(config=config, lang2id={}, speaker2id={})
    device = "cpu"
    if gpu:
        device = "cuda"
    model.to(device)
    for k, v in batch.items():
        if isinstance(v, torch.Tensor):
            batch[k] = v.to(device)
    if benchmark_type == BenchmarkType.training:
        benchmark_fn = model.forward
    else:
        benchmark_fn = partial(model.forward, inference=True)

    # INIT LOGGERS
    starter, ender = (
        torch.cuda.Event(enable_timing=True),
        torch.cuda.Event(enable_timing=True),
    )
    timings = np.zeros((repetitions, 1))
    # GPU-WARM-UP
    for _ in range(warmup_reps):
        _ = benchmark_fn(batch)
    # Forward
    for rep in range(repetitions):
        if gpu:
            starter.record()
            _ = benchmark_fn(batch)
            ender.record()
            # WAIT FOR GPU SYNC
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender)
            timings[rep] = curr_time
        else:
            t0 = time.time()
            _ = benchmark_fn(batch)
            t1 = time.time()
            curr_time = (t1 - t0) * 1000
            timings[rep] = curr_time
    print(
        f"Average forward pass for {benchmark_type.value} duration after {repetitions} repetitions: {np.sum(timings) / repetitions} ms Standard Deviation: {np.std(timings)}"
    )



import os
from pathlib import Path
from random import Random
from typing import Optional

import numpy as np
import pytorch_lightning as pl
import torch
from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils import (
    check_dataset_size,
    filter_dataset_based_on_target_text_representation_level,
    generic_psv_filelist_reader,
)
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
from torch.utils.data.sampler import Sampler

from .config import DFAlignerConfig


class AlignerDataModule(pl.LightningDataModule):
    def __init__(self, config: DFAlignerConfig):
        super().__init__()
        self.config = config
        self.train_sampler = None
        self.val_sampler = None
        self.batch_size = config.training.batch_size
        self.train_path = os.path.join(
            config.training.logger.save_dir,
            config.training.logger.name,
            "alignment_train_data.pth",
        )
        self.val_path = os.path.join(
            config.training.logger.save_dir,
            config.training.logger.name,
            "alignment_val_data.pth",
        )

        self.load_dataset()
        self.dataset_length = len(self.train_dataset) + len(self.val_dataset)

    def setup(self, stage: Optional[str] = None):
        # load it back here
        self.train_dataset = torch.load(self.train_path)
        self.val_dataset = torch.load(self.val_path)

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            pin_memory=False,
            sampler=self.train_sampler,
            collate_fn=collate_dataset,
            drop_last=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=1,
            pin_memory=False,
            sampler=self.val_sampler,
            collate_fn=collate_dataset,
            drop_last=True,
        )

    def predict_dataloader(self):
        return DataLoader(
            AlignerDataset(self.train_dataset + self.val_dataset, self.config),
            batch_size=self.batch_size,
            pin_memory=False,
            collate_fn=collate_dataset,
            drop_last=True,
        )

    def prepare_data(self):
        self.train_dataset = filter_dataset_based_on_target_text_representation_level(
            self.config.model.target_text_representation_level,
            self.train_dataset,
            "training",
            self.batch_size,
        )
        self.val_dataset = filter_dataset_based_on_target_text_representation_level(
            self.config.model.target_text_representation_level,
            self.val_dataset,
            "validation",
            self.batch_size,
        )
        train_samples = len(self.train_dataset)
        val_samples = len(self.val_dataset)
        check_dataset_size(self.batch_size, train_samples, "training")
        check_dataset_size(self.batch_size, val_samples, "validation")
        self.train_dataset = AlignerDataset(self.train_dataset, self.config)
        self.val_dataset = AlignerDataset(self.val_dataset, self.config)
        if self.config.training.binned_sampler:
            self.train_mel_lens = [d["mel_len"] for d in self.train_dataset]
            self.val_mel_lens = [d["mel_len"] for d in self.val_dataset]
            self.train_sampler = BinnedLengthSampler(
                mel_lens=self.train_mel_lens,
                batch_size=self.batch_size,
                bin_size=self.batch_size * 3,
                seed=self.config.preprocessing.dataset_split_seed,
            )
            self.val_sampler = BinnedLengthSampler(
                mel_lens=self.val_mel_lens,
                batch_size=self.batch_size,
                bin_size=self.batch_size * 3,
                seed=self.config.preprocessing.dataset_split_seed,
            )

        # save it to disk
        torch.save(self.train_dataset, self.train_path)
        torch.save(self.val_dataset, self.val_path)

    def load_dataset(self):
        # Can use same filelist as for feature prediction
        self.train_dataset = generic_psv_filelist_reader(
            self.config.training.training_filelist
        )
        self.val_dataset = generic_psv_filelist_reader(
            self.config.training.validation_filelist
        )


class AlignerDataset(Dataset):
    def __init__(self, data, config: DFAlignerConfig):
        super().__init__()
        self.config = config
        self.data = data
        self.preprocessed_dir = Path(self.config.preprocessing.save_dir)
        self.text_processor = TextProcessor(config.text)
        self.sep = "--"
        self.sampling_rate = self.config.preprocessing.audio.alignment_sampling_rate

    def _load_file(self, bn, spk, lang, dir, fn):
        return torch.load(
            self.preprocessed_dir / dir / self.sep.join([bn, spk, lang, fn])
        )

    def __getitem__(self, index):
        item = self.data[index]
        speaker = "default" if "speaker" not in item else item["speaker"]
        language = "default" if "language" not in item else item["language"]
        basename = item["basename"]
        mel = (
            self._load_file(
                basename,
                speaker,
                language,
                "spec",
                f"spec-{self.sampling_rate}-{self.config.preprocessing.audio.spec_type}.pt",
            )
            .squeeze()
            .transpose(0, 1)
        )  # [mel_bins, frames] -> [frames, mel_bins]
        match self.config.model.target_text_representation_level:
            case TargetTrainingTextRepresentationLevel.characters:
                text_tokens = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["character_tokens"]
                    )
                )
            case TargetTrainingTextRepresentationLevel.ipa_phones | TargetTrainingTextRepresentationLevel.phonological_features:
                text_tokens = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["phone_tokens"]
                    )
                )
            case _:
                raise NotImplementedError(
                    f"{self.config.model.target_text_representation_level} have not yet been implemented."
                )

        tokens_len = text_tokens.size(0)
        mel_len = mel.size(0)
        return {
            "basename": basename,
            "tokens": text_tokens,
            "mel": mel,
            "tokens_len": tokens_len,
            "mel_len": mel_len,
            "speaker": speaker,
            "language": language,
        }

    def __len__(self):
        return len(self.data)


# From https://github.com/fatchord/WaveRNN/blob/master/utils/dataset.py
class BinnedLengthSampler(Sampler):
    def __init__(self, mel_lens: torch.Tensor, batch_size: int, bin_size: int, seed=42):
        _, self.idx = torch.sort(torch.tensor(mel_lens))
        self.batch_size = batch_size
        self.bin_size = bin_size
        self.random = Random(seed)
        assert self.bin_size % self.batch_size == 0

    def __iter__(self):
        idx = self.idx.numpy()
        bins = []
        for i in range(len(idx) // self.bin_size):
            this_bin = idx[i * self.bin_size : (i + 1) * self.bin_size]
            self.random.shuffle(this_bin)
            bins += [this_bin]
        self.random.shuffle(bins)
        binned_idx = np.stack(bins).reshape(-1)
        if len(binned_idx) < len(idx):
            last_bin = idx[len(binned_idx) :]
            self.random.shuffle(last_bin)
            binned_idx = np.concatenate([binned_idx, last_bin])
        return iter(torch.IntTensor(binned_idx))

    def __len__(self):
        return len(self.idx)


def collate_dataset(batch: list[dict]) -> dict[str, torch.Tensor | list[str]]:
    tokens: torch.Tensor = pad_sequence(
        [b["tokens"] for b in batch], batch_first=True, padding_value=0
    )
    mels: torch.Tensor = pad_sequence(
        [b["mel"] for b in batch], batch_first=True, padding_value=0
    )
    tokens_len = torch.IntTensor([b["tokens_len"] for b in batch])
    mel_len = torch.IntTensor([b["mel_len"] for b in batch])
    return {
        "tokens": tokens,
        "mel": mels,
        "tokens_len": tokens_len,
        "mel_len": mel_len,
        "basename": [b["basename"] for b in batch],
        "language": [b["language"] for b in batch],
        "speaker": [b["speaker"] for b in batch],
    }



import sys

import numpy as np
import torch
import torch.nn.functional as F
from everyvoice.exceptions import BadDataError
from loguru import logger
from torch import nn
from torch.nn.utils.rnn import pad_sequence

from .attn.alignment import mas_width1
from .attn.attention import ConvAttention
from .config import FastSpeech2Config
from .layers import VarianceConvolutionLayer
from .type_definitions_heavy import InferenceControl, Stats, StatsInfo


class VariancePredictor(nn.Module):
    def __init__(
        self,
        input_dim: int,
        n_layers=5,
        n_channels=384,
        output_dim=1,
        kernel_size=5,
        dropout_rate=0.1,
        depthwise: bool = False,
    ):
        """Initialize VariancePredictor

        Args:
            input_dim (int): dimension of input
            n_layers (int, optional): number of layers. Defaults to 5.
            n_channels (int, optional): number of channels in convolutional layers. Defaults to 384.
            output_dim (int, optional): output dimension. Defaults to 2.
            kernel_size (int, optional): kernel size of convolution layers. Defaults to 5.
            dropout_rate (float, optional): dropout rate. Defaults to 0.1.
        """
        super().__init__()
        self.conv = nn.ModuleList()
        self.kernel_size = kernel_size
        for idx in range(n_layers):
            in_chans = input_dim if idx == 0 else n_channels
            self.conv.append(
                VarianceConvolutionLayer(
                    in_channels=in_chans,
                    out_channels=n_channels,
                    kernel_size=kernel_size,
                    dropout=dropout_rate,
                    depthwise=depthwise,
                )
            )
        self.linear = nn.Linear(n_channels, output_dim)

    def forward(self, x, mask=None):
        for m in self.conv:
            x = m(x)
        out = self.linear(x)
        out = out.squeeze(-1)
        if mask is not None:
            out = out * mask
        return out


class LengthRegulator(nn.Module):
    def forward(self, x, durations, max_length=None):
        repeated_list = [
            torch.repeat_interleave(x[i], durations[i], dim=0)
            for i in range(x.shape[0])
        ]
        lengths = torch.IntTensor([t.shape[0] for t in repeated_list])
        # FIXME: int(max_length) when max_length is None is invalid
        max_length = min(lengths.max(), int(max_length))
        mask = (
            torch.arange(max_length).expand(len(lengths), max_length)
            < lengths.unsqueeze(1)
        ).to(x.device)
        out = pad_sequence(repeated_list, batch_first=True, padding_value=0)
        if max_length is not None:
            out = out[:, :max_length]
        return out, mask


class VarianceAdaptor(nn.Module):
    """Variance Adaptor"""

    def __init__(self, config: FastSpeech2Config, stats: Stats):
        super().__init__()
        self.config = config
        self.stats = stats
        # Duration Predictor
        self.duration_predictor = VariancePredictor(
            input_dim=self.config.model.encoder.input_dim,
            n_layers=self.config.model.variance_predictors.duration.n_layers,
            n_channels=self.config.model.variance_predictors.duration.input_dim,
            output_dim=1,
            kernel_size=self.config.model.variance_predictors.duration.kernel_size,
            dropout_rate=self.config.model.variance_predictors.duration.dropout,
            depthwise=self.config.model.variance_predictors.duration.depthwise,
        )
        self.length_regulator = LengthRegulator()
        # Pitch Predictor
        self.pitch_predictor = VariancePredictor(
            input_dim=self.config.model.encoder.input_dim,
            n_layers=self.config.model.variance_predictors.pitch.n_layers,
            n_channels=self.config.model.variance_predictors.pitch.input_dim,
            output_dim=1,
            kernel_size=self.config.model.variance_predictors.pitch.kernel_size,
            dropout_rate=self.config.model.variance_predictors.pitch.dropout,
            depthwise=self.config.model.variance_predictors.pitch.depthwise,
        )
        self.pitch_embedding = nn.Embedding(
            self.config.model.variance_predictors.pitch.n_bins,
            self.config.model.variance_predictors.pitch.input_dim,
            # padding_idx=0,
        )
        self.pitch_bins = nn.Parameter(
            torch.linspace(
                self.stats.pitch.norm_min,
                self.stats.pitch.norm_max,
                self.config.model.variance_predictors.pitch.n_bins - 1,
            ),
            requires_grad=False,
        )
        # Energy Predictor
        self.energy_predictor = VariancePredictor(
            input_dim=self.config.model.encoder.input_dim,
            n_layers=self.config.model.variance_predictors.energy.n_layers,
            n_channels=self.config.model.variance_predictors.energy.input_dim,
            output_dim=1,
            kernel_size=self.config.model.variance_predictors.energy.kernel_size,
            dropout_rate=self.config.model.variance_predictors.energy.dropout,
            depthwise=self.config.model.variance_predictors.energy.depthwise,
        )
        self.energy_embedding = nn.Embedding(
            self.config.model.variance_predictors.energy.n_bins,
            self.config.model.variance_predictors.energy.input_dim,
            # padding_idx=0,
        )

        self.energy_bins = nn.Parameter(
            torch.linspace(
                self.stats.energy.norm_min,
                self.stats.energy.norm_max,
                self.config.model.variance_predictors.energy.n_bins - 1,
            ),
            requires_grad=False,
        )

        # Attention
        if self.config.model.learn_alignment:
            self.attention = ConvAttention(
                self.config.preprocessing.audio.n_mels,
                0,
                self.config.model.encoder.input_dim,
                use_query_proj=True,
                align_query_enc_type="3xconv",
            )

    def binarize_attention(self, attn, in_lens, out_lens):
        """For training purposes only. Binarizes attention with MAS.
           These will no longer recieve a gradient.
        Args:
            attn: B x 1 x max_mel_len x max_text_len
        """
        b_size = attn.shape[0]
        with torch.no_grad():
            attn_out_cpu = np.zeros(attn.data.shape, dtype=np.float32)
            log_attn_cpu = torch.log(attn.data).to(device="cpu", dtype=torch.float32)
            log_attn_cpu = log_attn_cpu.numpy()
            out_lens_cpu = out_lens.cpu()
            in_lens_cpu = in_lens.cpu()
            for ind in range(b_size):
                hard_attn = mas_width1(
                    log_attn_cpu[ind, 0, : out_lens_cpu[ind], : in_lens_cpu[ind]]
                )
                attn_out_cpu[
                    ind, 0, : out_lens_cpu[ind], : in_lens_cpu[ind]
                ] = hard_attn
            attn_out = torch.tensor(attn_out_cpu, device=attn.device, dtype=attn.dtype)
        return attn_out

    def get_variance_embedding(
        self,
        x,
        target,
        mask,
        predictor,
        embedding,
        stats: StatsInfo,
        bins,
        control,
        inference,
    ):
        prediction = predictor(x, mask)
        if not inference:
            buckets = torch.bucketize(target, bins)
            # max_b = max([max(b) for b in buckets])
            # min_b = min([min(b) for b in buckets])
            # breakpoint()
            embed = embedding(buckets.to(x.device))
        else:
            prediction = prediction * control
            embed = embedding(torch.bucketize(prediction, bins).to(x.device))
        return prediction, embed

    def average_variance(self, var, durs):
        durs_cums_ends = torch.cumsum(durs, dim=1).long()
        durs_cums_starts = F.pad(durs_cums_ends[:, :-1], (1, 0))
        var_nonzero_cums = F.pad(torch.cumsum(var != 0.0, dim=1), (1, 0))
        var_cums = F.pad(torch.cumsum(var, dim=1), (1, 0))

        var_sums = (
            torch.gather(var_cums, 1, durs_cums_ends)
            - torch.gather(var_cums, 1, durs_cums_starts)
        ).float()
        var_nelems = (
            torch.gather(var_nonzero_cums, 1, durs_cums_ends)
            - torch.gather(var_nonzero_cums, 1, durs_cums_starts)
        ).float()
        var_avg = torch.where(var_nelems == 0.0, var_nelems, var_sums / var_nelems)
        return var_avg

    def forward(
        self,
        text_emb,
        encoder_output,
        batch,
        src_mask,
        control=InferenceControl(),
        inference=False,
        teacher_forcing=False,
    ):  # sourcery skip: swap-if-expression
        # Get information from batch
        x = encoder_output.clone()
        energy_target = batch["energy"] if not inference else None
        pitch_target = batch["pitch"] if not inference else None
        duration_target = (
            batch["duration"] if batch["duration"][0] is not None else None
        )
        max_target_len = batch["max_mel_len"]
        src_mask = src_mask
        attn_logprob = None  # Overwritten if alignment is learned
        attn_soft = None
        attn_hard = None
        # speaker embedding is handled in main model
        # Alignment
        if (teacher_forcing or not inference) and self.config.model.learn_alignment:
            # make sure to do the alignments before folding
            attn_mask = src_mask[..., None] == 0
            # attn_mask should be 1 for unused timesteps in the text_enc_w_spkvec tensor
            attn_soft, attn_logprob = self.attention(
                batch["mel"].transpose(1, 2),
                text_emb.transpose(1, 2),
                batch["mel_lens"],
                attn_mask,
                key_lens=batch["src_lens"],
                keys_encoded=x,
                attn_prior=batch["duration"],
            )

            attn_hard = self.binarize_attention(
                attn_soft, batch["src_lens"], batch["mel_lens"]
            )

            # Viterbi --> durations
            attn_hard_dur = attn_hard.sum(2)[:, 0, :]
            duration_target = attn_hard_dur.int()
            if (
                pitch_target is not None and (pitch_target.size(1) == text_emb.size(1))
            ) or (
                energy_target is not None
                and (energy_target.size(1) == text_emb.size(1))
            ):
                logger.error(
                    "Your pitch and/or energy targets are already averaged across phones, but when you are learning alignment with phone-level energy or pitch modelling, you must have an un-averaged target for these as the duration of phones changes during training. This should happen automatically if you re-run the preprocessing step for energy and pitch."
                )
                sys.exit(1)
            if (
                energy_target is not None
                and self.config.model.variance_predictors.energy.level == "phone"
            ):
                energy_target = self.average_variance(energy_target, duration_target)
            if (
                pitch_target is not None
                and self.config.model.variance_predictors.pitch.level == "phone"
            ):
                pitch_target = self.average_variance(pitch_target, duration_target)
            try:
                equal_dur_targets = torch.eq(
                    duration_target.sum(dim=1), batch["mel_lens"]
                )
                assert torch.all(equal_dur_targets)
            except AssertionError as e:
                from itertools import compress

                mismatches = list(
                    compress(
                        batch["basename"], [not x for x in equal_dur_targets.tolist()]
                    )
                )
                raise BadDataError(
                    f"Something failed with the following items, please check them for errors: {mismatches}"
                ) from e
                sys.exit(1)
        # If phone-level variance, use src_mask before duration predictor and upsampling of x
        # using length regulator
        # otherwise use tgt_mask after upsampling
        if self.config.model.variance_predictors.energy.level == "phone":
            energy_prediction, energy_embedding = self.get_variance_embedding(
                x,
                energy_target,
                src_mask,
                self.energy_predictor,
                self.energy_embedding,
                self.stats.energy,
                self.energy_bins,
                control.energy,
                inference,
            )
            try:
                x = x + energy_embedding
            except RuntimeError as e:
                print(batch["basename"])
                print(f"energy target is size {energy_target.size()}")
                print(x.size())
                print(energy_embedding.size())
                breakpoint()
                raise e
        if self.config.model.variance_predictors.pitch.level == "phone":
            pitch_prediction, pitch_embedding = self.get_variance_embedding(
                x,
                pitch_target,
                src_mask,
                self.pitch_predictor,
                self.pitch_embedding,
                self.stats.pitch,
                self.pitch_bins,
                control.pitch,
                inference,
            )
            try:
                x = x + pitch_embedding
            except RuntimeError as e:
                print(x.size())
                print(batch["basename"])
                print(f"pitch target is size {pitch_target.size()}")
                print(pitch_embedding.size())
                breakpoint()
                raise e

        log_duration_prediction = self.duration_predictor(x, mask=src_mask)
        # upsampling from text time steps to mel time steps
        if teacher_forcing or not inference:
            x, tgt_mask = self.length_regulator(
                x, duration_target, max_length=max_target_len
            )
            duration_rounded = duration_target
        else:
            duration_rounded = torch.clamp(
                (
                    torch.round(torch.exp(log_duration_prediction) - 1)
                    * control.duration
                ),
                min=0,
            ).int()
            x, tgt_mask = self.length_regulator(
                x, duration_rounded, max_length=max_target_len
            )

        if self.config.model.variance_predictors.energy.level == "frame":
            energy_prediction, energy_embedding = self.get_variance_embedding(
                x,
                energy_target,
                tgt_mask,
                self.energy_predictor,
                self.energy_embedding,
                self.stats.energy,
                self.energy_bins,
                control.energy,
                inference,
            )
            x = x + energy_embedding

        if self.config.model.variance_predictors.pitch.level == "frame":
            pitch_prediction, pitch_embedding = self.get_variance_embedding(
                x,
                pitch_target,
                tgt_mask,
                self.pitch_predictor,
                self.pitch_embedding,
                self.stats.pitch,
                self.pitch_bins,
                control.pitch,
                inference,
            )
            x = x + pitch_embedding

        return {
            "output": x,
            "attn_logprob": attn_logprob,
            "attn_soft": attn_soft,
            "attn_hard": attn_hard,
            "duration_prediction": log_duration_prediction,
            "duration_target": duration_target,
            "pitch_prediction": pitch_prediction,
            "pitch_target": pitch_target,
            "energy_prediction": energy_prediction,
            "energy_target": energy_target,
            "duration_rounded": duration_rounded,
            "target_mask": tgt_mask,
        }



from enum import Enum
from pathlib import Path

import typer
from everyvoice.base_cli.interfaces import (
    complete_path,
    preprocess_base_command_interface,
    train_base_command_interface,
)
from loguru import logger
from merge_args import merge_args

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="A PyTorch Lightning implementation of the HiFiGAN and iSTFT-Net vocoders",
)


class PreprocessCategories(str, Enum):
    audio = "audio"
    spec = "spec"


@app.command()
@merge_args(preprocess_base_command_interface)
def preprocess(
    steps: list[PreprocessCategories] = typer.Option(
        [cat.value for cat in PreprocessCategories],
        "-s",
        "--steps",
        help="Which steps of the preprocessor to use. If none are provided, all steps will be performed.",
    ),
    **kwargs,
):
    from everyvoice.base_cli.helpers import preprocess_base_command

    from .config import HiFiGANConfig

    preprocess_base_command(
        model_config=HiFiGANConfig,
        steps=[step.name for step in steps],
        **kwargs,
    )


@app.command()
@merge_args(train_base_command_interface)
def train(**kwargs):
    from everyvoice.base_cli.helpers import train_base_command

    from .config import HiFiGANConfig
    from .dataset import HiFiGANDataModule
    from .model import HiFiGAN

    train_base_command(
        model_config=HiFiGANConfig,
        model=HiFiGAN,
        data_module=HiFiGANDataModule,
        monitor="validation/mel_spec_error",
        # We can't do this automatically with Lightning, so we do it manually in model.py
        gradient_clip_val=None,
        **kwargs,
    )


@app.command()
def synthesize(
    data_path: Path = typer.Option(
        ...,
        "--input",
        "-i",
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to a torch file containing time-oriented spectral features [T (frames), K (Mel bands)]",
        autocompletion=complete_path,
    ),
    generator_path: Path = typer.Option(
        ...,
        "--model",
        "-m",
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to a trained EveryVoice spec-to-wav model",
        autocompletion=complete_path,
    ),
):
    """Given some Mel spectrograms and a trained model, generate some audio. i.e. perform *copy synthesis*"""
    import torch
    from scipy.io.wavfile import write

    from .utils import load_hifigan_from_checkpoint, synthesize_data

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    checkpoint = torch.load(generator_path, map_location=device)
    data = torch.load(data_path, map_location=device)
    vocoder_model, vocoder_config = load_hifigan_from_checkpoint(checkpoint, device)
    wav, sr = synthesize_data(data, vocoder_model, vocoder_config)
    logger.info(f"Writing file {data_path}.wav")
    write(f"{data_path}.wav", sr, wav)


if __name__ == "__main__":
    app()



import builtins
import os
import string
import tempfile
from enum import Enum
from pathlib import Path
from types import MethodType
from typing import Callable, Iterable, NamedTuple, Optional, Sequence
from unittest import TestCase

import yaml
from anytree import RenderTree

# [Unit testing questionary](https://github.com/prompt-toolkit/python-prompt-toolkit/blob/master/docs/pages/advanced_topics/unit_testing.rst)
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput

from everyvoice.tests.stubs import (
    QuestionaryStub,
    Say,
    capture_stderr,
    capture_stdout,
    monkeypatch,
    null_patch,
    patch_menu_prompt,
)
from everyvoice.wizard import State, Step
from everyvoice.wizard import StepNames as SN
from everyvoice.wizard import Tour, basic, dataset, prompts
from everyvoice.wizard.basic import ConfigFormatStep
from everyvoice.wizard.main_tour import get_main_wizard_tour
from everyvoice.wizard.utils import EnumDict

CONTACT_INFO_STATE = State()
CONTACT_INFO_STATE[SN.contact_name_step.value] = "Test Name"
CONTACT_INFO_STATE[SN.contact_email_step.value] = "test@this.ca"


class RecursiveAnswers(NamedTuple):
    """Recursive answer for StepAndAnswer.children_answers, see StepAndAnswer
    documentation for a description of the fields here."""

    answer_or_monkey: Say | Callable
    children_answers: Optional[list["RecursiveAnswers"]] = None


class StepAndAnswer(NamedTuple):
    """named tuples to group together a step, its answer, and the answers of its children
    Members:
        step: an instance of a subclass of Step
        answer_or_monkey: either:
            - an instance of Say to get patched in for Step's prompt method
            or
            - a pre-instantiated monkeypatch context to use
        children_answers: an optional list of RecursiveAnswers to be used for
            the children of step this must align with what step.effect() will
            add as children given answer_or_monkey
    """

    step: Step
    answer_or_monkey: Say | Callable
    children_answers: Optional[list[RecursiveAnswers]] = None

    @property
    def monkey(self):
        """Return the monkey to use as a context manager"""
        if isinstance(self.answer_or_monkey, Say):
            return monkeypatch(self.step, "prompt", self.answer_or_monkey)
        else:
            return self.answer_or_monkey


def find_step(name: Enum, steps: Sequence[Step | list[Step]]):
    """Find a step with the given name in steps, of potentially variable depth"""
    for s in steps:
        if isinstance(s, list):
            try:
                return find_step(name, s)
            except IndexError:
                pass
        elif s.name == name.value:
            return s
    raise IndexError(f"Step {name} not found.")  # pragma: no cover


class WizardTest(TestCase):
    """Basic test for the configuration wizard"""

    data_dir = Path(__file__).parent / "data"

    def test_implementation_missing(self):
        nothing_step = Step(name="Dummy Step")
        no_validate_step = Step(name="Dummy Step", prompt_method=lambda: "test")
        no_prompt_step = Step(name="Dummy Step", validate_method=lambda: True)
        for step in [nothing_step, no_validate_step, no_prompt_step]:
            with self.assertRaises(NotImplementedError):
                step.run()

    def test_config_format_effect(self):
        """This is testing if a null key can be passed without throwing an
        error, as reported by marc tessier.  There are no assertions, it is just
        testing that no exceptions get raised.
        """
        config_step = basic.ConfigFormatStep(name="Config Step")
        self.assertTrue(config_step.validate("yaml"))
        self.assertTrue(config_step.validate("json"))
        with tempfile.TemporaryDirectory() as tmpdirname:
            config_step.state = State()
            config_step.state[SN.output_step.value] = tmpdirname
            config_step.state[SN.name_step.value] = config_step.name
            config_step.state.update(CONTACT_INFO_STATE)
            config_step.state["dataset_test"] = State()
            config_step.state["dataset_test"][SN.symbol_set_step.value] = {
                "characters": list(string.ascii_letters)
            }
            config_step.state["dataset_test"][SN.wavs_dir_step.value] = (
                Path(tmpdirname) / "test"
            )
            config_step.state["dataset_test"][SN.dataset_name_step.value] = "test"
            config_step.state["dataset_test"]["filelist_data"] = [
                {"basename": "0001", "text": "hello"},
                {"basename": "0002", "text": "hello", None: "test"},
                {"basename": "0003.wav", "text": "hello", None: "test"},
            ]
            config_step.state["dataset_test"]["sox_effects"] = []
            with capture_stdout() as stdout:
                config_step.effect()
            with open(Path(tmpdirname) / "Config Step" / "test-filelist.psv") as f:
                self.assertEqual(
                    f.read(), "basename|text\n0001|hello\n0002|hello\n0003|hello\n"
                )
            self.assertIn("Congratulations", stdout.getvalue())
            self.assertTrue(
                (Path(tmpdirname) / config_step.name / "logs_and_checkpoints").exists()
            )

    def test_access_response(self):
        root_step = Step(
            name="Dummy Step",
            prompt_method=lambda: "foo",
            validate_method=lambda x: True,
        )

        def validate(self, x):
            """Because the root step always returns 'foo', we only validate the second step if the prompt returns 'foobar'"""
            if self.parent.response + x == "foobar":
                return True

        second_step = Step(
            name="Dummy Step 2", prompt_method=lambda: "bar", parent=root_step
        )
        second_step.validate = MethodType(validate, second_step)
        for i, leaf in enumerate(RenderTree(root_step)):
            if i != 0:
                self.assertEqual(second_step.parent.response, "foo")
                self.assertTrue(leaf[2].validate("bar"))
                self.assertFalse(leaf[2].validate("foo"))
            leaf[2].run()

    def test_main_tour(self):
        tour = get_main_wizard_tour()
        self.assertGreater(len(tour.steps), 6)
        # TODO try to figure out how to actually run the tour in unit testing or
        # at least add more interesting assertions that just the fact that it's
        # got several steps.
        # self.monkey_run_tour() with a bunch of recursive answer would the thing to use here...

    def test_visualize(self):
        tour = get_main_wizard_tour()
        with capture_stdout() as out:
            tour.visualize()
        log = out.getvalue()
        self.assertIn("── Contact Name Step", log)
        self.assertIn("── Validate Wavs Step", log)

    def test_name_step(self):
        """Exercise providing a valid dataset name."""
        step = basic.NameStep()
        with capture_stdout() as stdout:
            with monkeypatch(builtins, "input", Say("myname")):
                step.run()
        self.assertEqual(step.response, "myname")
        self.assertIn("'myname'", stdout.getvalue())
        self.assertTrue(step.completed)

    def test_bad_name_step(self):
        """Exercise providing an invalid dataset name."""
        step = basic.NameStep("")
        # For unit testing, we cannot call step.run() if we patch a response
        # that will fail validation, because that would cause infinite
        # recursion.
        with capture_stdout() as stdout:
            self.assertFalse(step.validate("foo/bar"))
            self.assertFalse(step.validate(""))
        output = stdout.getvalue()
        self.assertIn("'foo/bar'", output)
        self.assertIn("is not valid", output)
        self.assertIn("your project needs a name", output)

        step = basic.NameStep("")
        with capture_stdout() as stdout:
            with monkeypatch(builtins, "input", Say(("bad/name", "good-name"), True)):
                step.run()
        output = stdout.getvalue()
        self.assertIn("'bad/name'", stdout.getvalue())
        self.assertIn("is not valid", stdout.getvalue())
        self.assertEqual(step.response, "good-name")

    def test_bad_contact_name_step(self):
        """Exercise providing an invalid contact name."""
        step = basic.ContactNameStep("")
        with capture_stdout() as stdout:
            self.assertFalse(step.validate("a"))
            self.assertFalse(step.validate(""))
        output = stdout.getvalue()
        self.assertIn("Sorry", output)
        self.assertIn("EveryVoice requires a name", output)

    def test_bad_contact_email_step(self):
        """Exercise providing an invalid contact email."""
        step = basic.ContactEmailStep("")
        with capture_stdout() as stdout:
            self.assertFalse(step.validate("test"))
            self.assertFalse(step.validate("test@"))
            self.assertFalse(step.validate("test@test."))
            self.assertTrue(step.validate("test@test.ca"))
            self.assertFalse(step.validate(""))
        output = stdout.getvalue()
        # Supporting email-validator prior and post 2.2.0 where the error string changed.
        self.assertTrue(
            "It must have exactly one @-sign" in output
            or "An email address must have an @-sign" in output
        )
        self.assertIn("There must be something after the @-sign", output)
        self.assertIn("An email address cannot end with a period", output)

    def test_no_permissions(self):
        """Exercise lacking permissions, then trying again"""
        tour = get_main_wizard_tour()
        permission_step = find_step(SN.dataset_permission_step, tour.steps)
        self.assertGreater(len(permission_step.children), 8)
        self.assertGreater(len(tour.root.descendants), 14)
        self.assertIn("dataset_0", tour.state)
        with patch_menu_prompt(0):  # 0 is no, I don't have permission
            permission_step.run()
        self.assertEqual(permission_step.children, ())
        self.assertLess(len(tour.root.descendants), 10)
        self.assertNotIn("dataset_0", tour.state)

        more_dataset_step = find_step(SN.more_datasets_step, tour.steps)
        with patch_menu_prompt(1):  # 1 is Yes, I have more data
            more_dataset_step.run()
        self.assertIn("dataset_1", tour.state)
        self.assertGreater(len(more_dataset_step.descendants), 8)
        self.assertGreater(len(tour.root.descendants), 14)

    def test_output_path_step(self):
        """Exercise the OutputPathStep"""
        tour = Tour(
            "testing",
            [
                basic.NameStep(),
                basic.OutputPathStep(),
            ],
        )

        # We need to answer the name step before we can validate the output path step
        step = tour.steps[0]
        with capture_stdout():
            with monkeypatch(builtins, "input", Say("myname")):
                step.run()

        step = tour.steps[1]
        with tempfile.TemporaryDirectory() as tmpdirname:
            file_path = os.path.join(tmpdirname, "exits-as-file")
            # Bad case 1: output dir exists and is a file
            with open(file_path, "w", encoding="utf8") as f:
                f.write("blah")
            with capture_stdout():
                self.assertFalse(step.validate(file_path))

            # Bad case 2: file called the same as the dataset exists in the output dir
            dataset_file = os.path.join(tmpdirname, "myname")
            with open(dataset_file, "w", encoding="utf8") as f:
                f.write("blah")
            with capture_stdout():
                self.assertFalse(step.validate(tmpdirname))
            os.unlink(dataset_file)

            # Bad case 3: file under read-only directory
            ro_dir = Path(tmpdirname) / "read-only"
            ro_dir.mkdir(mode=0x555)
            with capture_stdout() as out:
                self.assertFalse(step.validate(str(ro_dir)))
            self.assertIn("could not create", out.getvalue())

            # Good case
            with capture_stdout() as stdout:
                with monkeypatch(step, "prompt", Say(tmpdirname)):
                    step.run()
            self.assertIn("will put your files", stdout.getvalue())

    def test_more_data_step(self):
        """Exercise giving an invalid response and a yes response to more data."""
        tour = Tour(
            "testing",
            [dataset.FilelistStep(state_subset="dataset_0"), basic.MoreDatasetsStep()],
        )

        step = tour.steps[1]
        self.assertFalse(step.validate("foo"))
        self.assertTrue(step.validate("yes"))
        self.assertEqual(len(step.children), 0)

        with patch_menu_prompt(0):  # answer 0 is "no"
            step.run()
        self.assertEqual(len(step.children), 1)
        self.assertIsInstance(step.children[0], basic.ConfigFormatStep)

        with patch_menu_prompt(1):  # answer 1 is "yes"
            step.run()
        self.assertGreater(len(step.descendants), 10)

    def test_no_data_to_save(self):
        """When the tour created no datasets at all, saving the config is skipped."""
        tour = Tour("testing", [basic.MoreDatasetsStep()])
        step = tour.steps[0]
        with patch_menu_prompt(0), capture_stdout() as out:  # answer 0 is "no"
            step.run()
        self.assertEqual(len(step.children), 0)
        self.assertIn("No dataset to save", out.getvalue())

    def test_dataset_name(self):
        step = dataset.DatasetNameStep()
        with monkeypatch(builtins, "input", Say(("", "bad/name", "good-name"), True)):
            with capture_stdout() as stdout:
                step.run()
        output = stdout.getvalue().split("\n")
        self.assertIn("your dataset needs a name", output[0])
        self.assertIn("is not valid", output[1])
        self.assertIn("finished the configuration", output[2])
        self.assertTrue(step.completed)

    def test_wavs_dir(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            no_wavs_dir = os.path.join(tmpdirname, "no-wavs-here")
            os.mkdir(no_wavs_dir)

            has_wavs_dir = os.path.join(tmpdirname, "there-are-wavs-here")
            os.mkdir(has_wavs_dir)
            with open(os.path.join(has_wavs_dir, "foo.wav"), "wb") as f:
                f.write(b"A fantastic sounding clip! (or not...)")

            step = dataset.WavsDirStep("")
            with monkeypatch(
                dataset,
                "questionary",
                QuestionaryStub(("not-a-path", no_wavs_dir, has_wavs_dir)),
            ):
                with capture_stdout():
                    step.run()
            self.assertTrue(step.completed)
            self.assertEqual(step.response, has_wavs_dir)

    def test_sample_rate_config(self):
        step = dataset.SampleRateConfigStep("")
        with monkeypatch(
            dataset,
            "questionary",
            QuestionaryStub(
                (
                    "not an int",  # obvious not valid
                    "",  # ditto
                    3.1415,  # floats are also not allowed
                    50,  # this is below the minimum 100 allowed
                    512,  # yay, a good response!
                    1024,  # won't get used becasue 512 is good.
                )
            ),
        ):
            with capture_stdout() as stdout:
                step.run()
        output = stdout.getvalue().split("\n")
        for i in range(4):
            self.assertIn("not a valid sample rate", output[i])
        self.assertTrue(step.completed)
        self.assertEqual(step.response, 512)

    def test_dataset_subtour(self):
        tour = Tour("unit testing", steps=dataset.get_dataset_steps())

        filelist = str(self.data_dir / "unit-test-case1.psv")
        filelist_step = find_step(SN.filelist_step, tour.steps)
        monkey = monkeypatch(filelist_step, "prompt", Say(filelist))
        with monkey:
            filelist_step.run()

        permission_step = find_step(SN.dataset_permission_step, tour.steps)
        with patch_menu_prompt(1):  # 1 is "yes, I have permission"
            permission_step.run()
        self.assertTrue(
            permission_step.state[SN.dataset_permission_step].startswith("Yes")
        )

        format_step = find_step(SN.filelist_format_step, tour.steps)
        with patch_menu_prompt(0):  # 0 is "psv"
            format_step.run()
        self.assertEqual(len(format_step.children), 3)

        step = format_step.children[0]
        self.assertIsInstance(step, dataset.HasHeaderLineStep)
        self.assertEqual(step.name, SN.data_has_header_line_step.value)
        with patch_menu_prompt(1):  # 1 is "yes"
            step.run()
        self.assertEqual(step.state[SN.data_has_header_line_step.value], "yes")
        self.assertEqual(len(step.state["filelist_data_list"]), 4)

        step = format_step.children[1]
        self.assertIsInstance(step, dataset.HeaderStep)
        self.assertEqual(step.name, SN.basename_header_step.value)
        with patch_menu_prompt(1):  # 1 is second column
            step.run()
        self.assertEqual(step.response, 1)
        self.assertEqual(step.state["filelist_headers"][1], "basename")

        step = format_step.children[2]
        self.assertIsInstance(step, dataset.HeaderStep)
        self.assertEqual(step.name, SN.text_header_step.value)
        with patch_menu_prompt(1):  # 1 is second remaining column, i.e., third column
            step.run()
        self.assertEqual(step.state["filelist_headers"][2], "text")

        text_representation_step = find_step(
            SN.filelist_text_representation_step, tour.steps
        )
        with patch_menu_prompt(0):  # 0 is "characters"
            text_representation_step.run()
        self.assertEqual(step.state["filelist_headers"][2], "characters")
        speaker_step = find_step(SN.data_has_speaker_value_step, tour.steps)
        children_before = len(speaker_step.children)
        with patch_menu_prompt(0):  # 0 is "no"
            speaker_step.run()
        self.assertEqual(len(speaker_step.children), children_before)

        language_step = find_step(SN.data_has_language_value_step, tour.steps)
        children_before = len(language_step.children)
        with patch_menu_prompt(0):  # 0 is "no"
            language_step.run()
        self.assertEqual(len(language_step.children), children_before + 1)
        self.assertIsInstance(language_step.children[0], dataset.SelectLanguageStep)

        select_lang_step = language_step.children[0]
        with capture_stdout(), capture_stderr():
            with patch_menu_prompt(15):  # some arbitrary language from the list
                select_lang_step.run()
        # print(select_lang_step.state)
        self.assertEqual(
            select_lang_step.state["filelist_headers"],
            ["unknown_0", "basename", "characters", "unknown_3"],
        )

        wavs_dir_step = find_step(SN.wavs_dir_step, tour.steps)
        with monkeypatch(wavs_dir_step, "prompt", Say(str(self.data_dir))):
            wavs_dir_step.run()

        validate_wavs_step = find_step(SN.validate_wavs_step, tour.steps)
        with patch_menu_prompt(1), capture_stdout() as out:
            validate_wavs_step.run()
        self.assertEqual(step.state[SN.validate_wavs_step][:2], "No")
        self.assertIn("Warning: 3 wav files were not found", out.getvalue())

        text_processing_step = find_step(SN.text_processing_step, tour.steps)
        # 0 is lowercase, 1 is NFC Normalization, select both
        with monkeypatch(dataset, "tqdm", lambda seq, desc: seq):
            with patch_menu_prompt([0, 1]):
                text_processing_step.run()
        # print(text_processing_step.state)
        self.assertEqual(
            text_processing_step.state["filelist_data_list"][2][2],
            "cased \t nfd: éàê nfc: éàê",  # the "nfd: éàê" bit here is now NFC
        )

        # Make sure realoading the data as dict stripped the header line
        self.assertEqual(len(step.state["filelist_data"]), 3)

        sox_effects_step = find_step(SN.sox_effects_step, tour.steps)
        # 0 is resample to 22050 kHz, 2 is remove silence at start and end
        with patch_menu_prompt([0, 1]):
            sox_effects_step.run()
        # print(sox_effects_step.state["sox_effects"])
        self.assertEqual(
            sox_effects_step.state["sox_effects"],
            [
                ["channels", "1"],
                ["norm", "-3.0"],
                [
                    "silence",
                    "1",
                    "0.1",
                    "0.1%",
                ],
                ["reverse"],
                [
                    "silence",
                    "1",
                    "0.1",
                    "0.1%",
                ],
                ["reverse"],
            ],
        )

        symbol_set_step = find_step(SN.symbol_set_step, tour.steps)
        self.assertEqual(len(symbol_set_step.state["filelist_data"]), 3)
        with capture_stdout(), capture_stderr():
            symbol_set_step.run()
        self.assertEqual(len(symbol_set_step.state[SN.symbol_set_step.value]), 2)
        self.assertIn("t͡s", symbol_set_step.state[SN.symbol_set_step.value]["phones"])

    def test_wrong_fileformat_psv(self):
        tour = Tour(
            name="mismatched fileformat",
            steps=[
                dataset.FilelistStep(),
                dataset.FilelistFormatStep(),
            ],
        )
        filelist = str(self.data_dir / "unit-test-case2.psv")

        filelist_step = tour.steps[0]
        with monkeypatch(filelist_step, "prompt", Say(filelist)):
            filelist_step.run()
        format_step = tour.steps[1]
        # try with: 1/tsv (wrong), 2/csv (wrong), 3/festival (wrong) and finally 0 tsv (right)
        with patch_menu_prompt((1, 2, 3, 0), multi=True) as stdout:
            format_step.run()
        output = stdout.getvalue()
        self.assertIn("does not look like a 'tsv'", output)
        self.assertIn("does not look like a 'csv'", output)
        self.assertIn("is not in the festival format", output)
        self.assertTrue(format_step.completed)
        # print(format_step.state)

    def test_wrong_fileformat_festival(self):
        tour = Tour(
            name="mismatched fileformat",
            steps=[
                dataset.FilelistStep(),
                dataset.FilelistFormatStep(),
            ],
        )
        filelist = str(self.data_dir / "unit-test-case3.festival")

        filelist_step = tour.steps[0]
        with monkeypatch(filelist_step, "prompt", Say(filelist)):
            filelist_step.run()
        format_step = tour.steps[1]
        # try with: 0/psv (wrong), 1/tsv (wrong), 2/csv (wrong), and finally 3/festival (right)
        with patch_menu_prompt((0, 1, 2, 3), multi=True) as stdout:
            format_step.run()
        output = stdout.getvalue()
        self.assertIn("does not look like a 'psv'", output)
        self.assertIn("does not look like a 'tsv'", output)
        self.assertIn("does not look like a 'csv'", output)
        self.assertTrue(format_step.completed)
        # print(format_step.state)

    def test_validate_path(self):
        """Unit testing for validate_path() in isolation."""
        from everyvoice.wizard.validators import validate_path

        with self.assertRaises(ValueError):
            validate_path("", is_dir=False, is_file=False)
        with self.assertRaises(ValueError):
            validate_path("", is_dir=True, is_file=True)
        with self.assertRaises(ValueError):
            validate_path("")
        with tempfile.TemporaryDirectory() as tmpdirname, capture_stdout():
            self.assertTrue(
                validate_path(tmpdirname, is_dir=True, is_file=False, exists=True)
            )
            self.assertFalse(
                validate_path(tmpdirname, is_dir=False, is_file=True, exists=True)
            )
            self.assertFalse(
                validate_path(tmpdirname, is_dir=True, is_file=False, exists=False)
            )
            file_name = os.path.join(tmpdirname, "some-file-name")
            with open(file_name, "w", encoding="utf8") as f:
                f.write("foo")
            self.assertTrue(
                validate_path(file_name, is_dir=False, is_file=True, exists=True)
            )
            self.assertFalse(
                validate_path(file_name, is_dir=False, is_file=True, exists=False)
            )

            not_file_name = os.path.join(tmpdirname, "file-does-not-exist")
            self.assertFalse(
                validate_path(not_file_name, is_dir=False, is_file=True, exists=True)
            )
            self.assertTrue(
                validate_path(not_file_name, is_dir=False, is_file=True, exists=False)
            )

    def test_prompt(self):
        with patch_menu_prompt(0):
            answer = prompts.get_response_from_menu_prompt(
                choices=("choice1", "choice2")
            )
            self.assertEqual(answer, "choice1")
        with patch_menu_prompt(1) as stdout:
            answer = prompts.get_response_from_menu_prompt(
                "some question", ("choice1", "choice2")
            )
            self.assertEqual(answer, "choice2")
            self.assertIn("some question", stdout.getvalue())
        with patch_menu_prompt((2, 4)):
            answer = prompts.get_response_from_menu_prompt(
                choices=("a", "b", "c", "d", "e"), multi=True
            )
            self.assertEqual(answer, ["c", "e"])
        with patch_menu_prompt(1):
            answer = prompts.get_response_from_menu_prompt(
                choices=("a", "b", "c", "d", "e"), return_indices=True
            )
            self.assertEqual(answer, 1)

    def monkey_run_tour(
        self, name: str, steps_and_answers: list[StepAndAnswer]
    ) -> tuple[Tour, str]:
        """Create and run a tour with the monkey answers given

        Args:
            name: Name to give the tour when creating it
            steps_and_answers: a list of steps, answers, and optional recursive answers
                (Step, Answer/Monkey, optional [children answers/monkeys])
                where
                 - Step is an instantiated subclass of Step
                 - either:
                   - Answer is an instance of Say to get patched in for Step's prompt method
                   or
                   - Monkey is an instantiated monkeypatch context to use
                 - [children answers/monkeys] is an optional list of tuples
                   (Answer/Monkey, optional recursive [chidren answer/monkeys])
                   to be used recursively for the children of Step.
                   This must align with what Step.effect() adds as children.

        Returns: (tour, logs_from_stdout)
        """

        def recursive_helper(steps_and_answers: Iterable[StepAndAnswer]):
            """Run all the steps with the patched answers, recursively running children
            steps that get added, if requested by the specification of children_answers.
            """
            for step_and_answer in steps_and_answers:
                step = step_and_answer.step
                # print(step.name)
                old_children_len = len(step.children)
                with step_and_answer.monkey:
                    step.run()
                if (
                    len(step.children) > old_children_len
                    and step_and_answer.children_answers
                ):
                    # Here we assemble steps_and_answers for the recursive call from
                    # the actual children of step and the provided children_answers.
                    recursive_helper(
                        steps_and_answers=[
                            StepAndAnswer(
                                child_step,
                                answer_or_monkey=recursive_answers.answer_or_monkey,
                                children_answers=recursive_answers.children_answers,
                            )
                            for child_step, recursive_answers in zip(
                                step.children,
                                step_and_answer.children_answers,
                            )
                        ]
                    )

        tour = Tour(name, steps=[step for (step, *_) in steps_and_answers])
        # fail on accidentally shared initializer
        self.assertTrue(tour.state == {} or tour.state == {"dataset_0": {}})
        with capture_stdout() as out, capture_stderr():
            recursive_helper(steps_and_answers)
        return tour, out.getvalue()

    def test_monkey_tour_1(self):
        with tempfile.TemporaryDirectory() as tmpdirname, capture_stdout():
            tour, _ = self.monkey_run_tour(
                "monkey tour 1",
                [
                    StepAndAnswer(basic.NameStep(), Say("my-dataset-name")),
                    StepAndAnswer(basic.OutputPathStep(), Say(str(tmpdirname))),
                ],
            )
        self.assertEqual(tour.state[SN.name_step.value], "my-dataset-name")
        self.assertEqual(tour.state[SN.output_step.value], tmpdirname)

    def test_monkey_tour_2(self):
        data_dir = Path(__file__).parent / "data"
        tour, out = self.monkey_run_tour(
            "monkey tour 2",
            [
                StepAndAnswer(
                    dataset.FilelistStep(),
                    Say(str(data_dir / "metadata.psv")),
                ),
                StepAndAnswer(dataset.FilelistFormatStep(), Say("psv")),
                StepAndAnswer(
                    dataset.FilelistTextRepresentationStep(), Say("characters")
                ),
                StepAndAnswer(
                    dataset.HasSpeakerStep(),
                    Say("yes"),
                    children_answers=[RecursiveAnswers(Say(3))],
                ),
                StepAndAnswer(
                    dataset.HasLanguageStep(),
                    Say("no"),
                    children_answers=[RecursiveAnswers(Say("eng"))],
                ),
                StepAndAnswer(dataset.WavsDirStep(), Say(str(data_dir))),
                StepAndAnswer(
                    dataset.ValidateWavsStep(),
                    patch_menu_prompt(0),  # 0 is Yes
                    children_answers=[
                        RecursiveAnswers(Say(str(data_dir / "lj/wavs"))),
                        RecursiveAnswers(null_patch()),
                    ],
                ),
                StepAndAnswer(dataset.TextProcessingStep(), Say([0, 1])),
                StepAndAnswer(
                    dataset.SymbolSetStep(),
                    Say(True),
                ),
                StepAndAnswer(dataset.SoxEffectsStep(), Say([0])),
                StepAndAnswer(dataset.DatasetNameStep(), Say("my-monkey-dataset")),
            ],
        )

        tree = str(RenderTree(tour.root))
        self.assertIn("├── Validate Wavs Step", tree)
        self.assertIn("│   └── Validate Wavs Step", tree)
        self.assertIn("Great! All audio files found in directory", out)

        # print(tour.state)
        self.assertEqual(len(tour.state["filelist_data"]), 5)
        self.assertTrue(tour.steps[-1].completed)

    def test_get_iso_code(self):
        self.assertEqual(dataset.get_iso_code("eng"), "eng")
        self.assertEqual(dataset.get_iso_code("[eng]"), "eng")
        self.assertEqual(dataset.get_iso_code("es"), "es")
        self.assertEqual(dataset.get_iso_code("[es]"), "es")
        self.assertIs(dataset.get_iso_code(None), None)

    def test_with_language_column(self):
        data_dir = Path(__file__).parent / "data"
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            tour, _ = self.monkey_run_tour(
                "tour with language column",
                [
                    StepAndAnswer(basic.NameStep(), Say("project")),
                    StepAndAnswer(basic.ContactNameStep(), Say("Test Name")),
                    StepAndAnswer(basic.ContactEmailStep(), Say("info@everyvoice.ca")),
                    StepAndAnswer(basic.OutputPathStep(), Say(str(tmpdir / "out"))),
                    StepAndAnswer(
                        dataset.WavsDirStep(state_subset="dataset_0"),
                        Say(str(data_dir)),
                    ),
                    StepAndAnswer(
                        dataset.FilelistStep(state_subset="dataset_0"),
                        Say(str(data_dir / "language-col.tsv")),
                    ),
                    StepAndAnswer(
                        dataset.FilelistFormatStep(state_subset="dataset_0"), Say("tsv")
                    ),
                    StepAndAnswer(
                        dataset.FilelistTextRepresentationStep(
                            state_subset="dataset_0"
                        ),
                        Say("characters"),
                    ),
                    StepAndAnswer(
                        dataset.TextProcessingStep(state_subset="dataset_0"),
                        Say([0, 1]),
                    ),
                    StepAndAnswer(
                        dataset.HasSpeakerStep(state_subset="dataset_0"),
                        Say("yes"),
                        children_answers=[RecursiveAnswers(Say(2))],
                    ),
                    StepAndAnswer(
                        dataset.HasLanguageStep(state_subset="dataset_0"),
                        Say("yes"),
                        children_answers=[RecursiveAnswers(Say(3))],
                    ),
                    StepAndAnswer(
                        dataset.SymbolSetStep(state_subset="dataset_0"),
                        Say(True),
                    ),
                    StepAndAnswer(
                        dataset.SoxEffectsStep(state_subset="dataset_0"), Say([0])
                    ),
                    StepAndAnswer(
                        dataset.DatasetNameStep(state_subset="dataset_0"),
                        Say("my-monkey-dataset"),
                    ),
                    StepAndAnswer(
                        basic.MoreDatasetsStep(),
                        Say("no"),
                        children_answers=[RecursiveAnswers(Say("yaml"))],
                    ),
                ],
            )

            self.assertEqual(tour.state["dataset_0"][SN.speaker_header_step.value], 2)
            self.assertEqual(tour.state["dataset_0"][SN.language_header_step.value], 3)
            self.assertTrue(tour.steps[-1].completed)

            with open(tmpdir / "out/project/config/everyvoice-text-to-spec.yaml") as f:
                text_to_spec_config = "\n".join(f)
            self.assertIn("multilingual: true", text_to_spec_config)
            self.assertIn("multispeaker: true", text_to_spec_config)

    def test_no_header_line(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            with open(tmpdir / "filelist.psv", "w", encoding="utf8") as f:
                f.write("f1|foo bar|Joe\nf2|bar baz|Joe\nf3|baz foo|Joe\n")
            for basename in ("f1", "f2", "f3"):
                with open(tmpdir / (basename + ".wav"), "wb"):
                    pass
            tour, _ = self.monkey_run_tour(
                "Tour with datafile missing the header line",
                [
                    StepAndAnswer(basic.NameStep(), Say("project")),
                    StepAndAnswer(basic.ContactNameStep(), Say("Test Name")),
                    StepAndAnswer(basic.ContactEmailStep(), Say("info@everyvoice.ca")),
                    StepAndAnswer(basic.OutputPathStep(), Say(str(tmpdir / "out"))),
                    StepAndAnswer(
                        dataset.WavsDirStep(state_subset="dataset_0"), Say(str(tmpdir))
                    ),
                    StepAndAnswer(
                        dataset.FilelistStep(state_subset="dataset_0"),
                        Say(str(tmpdir / "filelist.psv")),
                    ),
                    StepAndAnswer(
                        dataset.FilelistFormatStep(state_subset="dataset_0"),
                        Say("psv"),
                        children_answers=[
                            RecursiveAnswers(Say("no")),  # no header line
                            RecursiveAnswers(Say(0)),  # column 0 is basename
                            RecursiveAnswers(Say(1)),  # column 1 is text
                        ],
                    ),
                    StepAndAnswer(
                        dataset.FilelistTextRepresentationStep(
                            state_subset="dataset_0"
                        ),
                        Say("characters"),
                    ),
                    StepAndAnswer(
                        dataset.HasSpeakerStep(state_subset="dataset_0"),
                        Say("no"),
                    ),
                    StepAndAnswer(
                        dataset.HasLanguageStep(state_subset="dataset_0"),
                        Say("no"),
                        children_answers=[RecursiveAnswers(Say("und"))],
                    ),
                    StepAndAnswer(
                        dataset.TextProcessingStep(state_subset="dataset_0"),
                        Say(()),
                    ),
                    StepAndAnswer(
                        dataset.SymbolSetStep(state_subset="dataset_0"),
                        Say(True),
                    ),
                    StepAndAnswer(
                        dataset.SoxEffectsStep(state_subset="dataset_0"),
                        Say([]),
                    ),
                    StepAndAnswer(
                        dataset.DatasetNameStep(state_subset="dataset_0"),
                        Say("dataset"),
                    ),
                    StepAndAnswer(
                        basic.MoreDatasetsStep(),
                        Say("no"),
                        children_answers=[RecursiveAnswers(Say("yaml"))],
                    ),
                ],
            )
            self.assertEqual(len(tour.state["dataset_0"]["filelist_data"]), 3)
            with open(tmpdir / "out/project/dataset-filelist.psv") as f:
                output_filelist = list(f)
            self.assertEqual(len(output_filelist), 4)
            with open(tmpdir / "out/project/config/everyvoice-text-to-spec.yaml") as f:
                text_to_spec_config = "\n".join(f)
            self.assertIn("multilingual: false", text_to_spec_config)
            self.assertIn("multispeaker: false", text_to_spec_config)

    def test_running_out_of_columns(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            with open(tmpdir / "filelist.psv", "w", encoding="utf8") as f:
                f.write("basename|text\nf1|foo bar\nf2|bar baz\nf3|baz foo\n")
            for basename in ("f1", "f2", "f3"):
                with open(tmpdir / (basename + ".wav"), "wb"):
                    pass
            tour, _ = self.monkey_run_tour(
                "Tour without enough columns to have speaker or language",
                [
                    StepAndAnswer(basic.NameStep(), Say("project")),
                    StepAndAnswer(basic.ContactNameStep(), Say("Test Name")),
                    StepAndAnswer(basic.ContactEmailStep(), Say("info@everyvoice.ca")),
                    StepAndAnswer(basic.OutputPathStep(), Say(str(tmpdir / "out"))),
                    StepAndAnswer(dataset.WavsDirStep(), Say(str(tmpdir))),
                    StepAndAnswer(
                        dataset.FilelistStep(),
                        Say(str(tmpdir / "filelist.psv")),
                    ),
                    StepAndAnswer(
                        dataset.FilelistFormatStep(),
                        Say("psv"),
                    ),
                    StepAndAnswer(
                        dataset.FilelistTextRepresentationStep(), Say("characters")
                    ),
                    StepAndAnswer(
                        dataset.HasSpeakerStep(),
                        patch_menu_prompt(1),
                        children_answers=[RecursiveAnswers(Say("foo"))],
                    ),
                    StepAndAnswer(
                        dataset.HasLanguageStep(),
                        patch_menu_prompt(1),
                        children_answers=[RecursiveAnswers(Say("bar"))],
                    ),
                    StepAndAnswer(dataset.SelectLanguageStep(), Say("und")),
                    StepAndAnswer(dataset.DatasetNameStep(), Say("dataset")),
                    StepAndAnswer(
                        basic.MoreDatasetsStep(),
                        Say("no"),
                        children_answers=[RecursiveAnswers(Say("yaml"))],
                    ),
                ],
            )
            self.assertEqual(tour.state["filelist_headers"], ["basename", "characters"])
            self.assertEqual(tour.state[SN.data_has_speaker_value_step.value], "no")
            self.assertEqual(tour.state[SN.data_has_language_value_step.value], "no")

    def test_keyboard_interrupt(self):
        step = basic.NameStep()
        with self.assertRaises(KeyboardInterrupt):
            with monkeypatch(builtins, "input", Say(KeyboardInterrupt())):
                step.run()

        step = dataset.WavsDirStep()
        with self.assertRaises(KeyboardInterrupt):
            with monkeypatch(
                dataset, "questionary", QuestionaryStub([KeyboardInterrupt()])
            ):
                step.run()

        step = basic.MoreDatasetsStep()
        with self.assertRaises(KeyboardInterrupt):
            with patch_menu_prompt(KeyboardInterrupt()):
                step.run()

    def test_give_up_after_twenty_failures(self):
        step = dataset.WavsDirStep()
        with monkeypatch(step, "prompt", Say("no/such/directory")):
            with capture_stdout(), self.assertRaises(SystemExit):
                step.run()

    def test_leading_white_space_in_outpath(self):
        """
        Make sure we strip leading spaces when the user accidentally adds a
        leading space when specifying an output path.
        """
        with capture_stdout(), tempfile.TemporaryDirectory() as tmpdirname:
            with create_pipe_input() as pipe_input:
                # NOTE: we use `` to replace the `.` with our path with leading spaces.
                pipe_input.send_text(f"  {tmpdirname}\n")
                with create_app_session(input=pipe_input, output=DummyOutput()):
                    tour = Tour(
                        name="trimming leading spaces",
                        steps=[
                            basic.OutputPathStep(),
                        ],
                        state={SN.name_step.value: "output_dir_with_leading_spaces"},
                    )
                    tour.run()
        self.assertFalse(tour.state[SN.output_step.value].startswith(" "))
        self.assertEqual(tour.state[SN.output_step.value], tmpdirname)

    def test_leading_white_space_in_wav_dir(self):
        """
        Make sure we strip leading spaces when the user accidentally adds a
        leading space when specifying a wav directory.
        """
        step = dataset.WavsDirStep()
        path = Path(__file__).parent
        with create_pipe_input() as pipe_input:
            pipe_input.send_text(f" {path}\n")
            with create_app_session(input=pipe_input, output=DummyOutput()):
                step.run()
        self.assertFalse(step.response.startswith(" "))
        self.assertEqual(step.response, str(path))

    def test_leading_white_space_in_filelist(self):
        """
        Make sure we strip leading spaces when the user accidentally adds a
        leading space when specifying a filelist.
        """
        step = dataset.FilelistStep()
        path = Path(__file__).parent / "data/unit-test-case1.psv"
        with create_pipe_input() as pipe_input:
            pipe_input.send_text(f" {path}\n")
            with create_app_session(input=pipe_input, output=DummyOutput()):
                step.run()
        self.assertFalse(step.response.startswith(" "))
        self.assertEqual(step.response, str(path))


class WavFileDirectoryRelativePathTest(TestCase):
    """
    Make sure the wav files directory path is correctly handle when transformed
    to a relative path.
    """

    data_dir = Path(__file__).parent / "data"

    def setUp(self):
        """
        Create a mock state instead of doing all prior steps to ConfigFormatStep.
        """
        state = State(
            {
                SN.output_step.value: "John/Smith",
                SN.name_step.value: "Unittest",
                "dataset_0": State(
                    {
                        SN.dataset_name_step.value: "unit",
                        SN.wavs_dir_step.value: "Common-Voice",
                        SN.symbol_set_step.value: {
                            "characters": [
                                " ",
                                ",",
                                ".",
                                "A",
                                "D",
                                "E",
                                "H",
                                "I",
                                "J",
                                "K",
                            ]
                        },
                        "filelist_data": [
                            {
                                "text": "Sentence 1",
                                "basename": "5061f5c3-3bf9-42c6-a268-435c146efaf6/dd50ed81b889047cb4399e34b650a91fcbd3b2a5e36cf0068251d64274bffb61",
                                "language": "und",
                            },
                            {
                                "text": "Sentence 2",
                                "basename": "5061f5c3-3bf9-42c6-a268-435c146efaf6/6c45ab8c6e2454142c95319ca37f7e4ff6526dddbcc7fc540572e4e53264ec47",
                                "language": "und",
                            },
                            {
                                "text": "Sentence 3",
                                "basename": "5061f5c3-3bf9-42c6-a268-435c146efaf6/3947ae033faeb793e00f836648e240bc91c821798bccc76656ad3e7030b38878",
                                "language": "und",
                            },
                            {
                                "text": "Sentence 4",
                                "basename": "5061f5c3-3bf9-42c6-a268-435c146efaf6/65b61440f9621084a1a1d8c461d177c765fad3aff91e0077296081931929629b",
                                "language": "und",
                            },
                            {
                                "text": "Sentence 5",
                                "basename": "5061f5c3-3bf9-42c6-a268-435c146efaf6/8a124117481eaf8f91d23aa3acda301e7fae7de85e98c016383381d54a3d5049",
                                "language": "und",
                            },
                        ],
                        "sox_effects": [["channel", "1"]],
                    }
                ),
            }
        )
        self.config = ConfigFormatStep()
        self.config.response = "yaml"
        self.config.state = state

    def test_wav_file_directory_local(self):
        """
        output directory is `.`
        wav files directory located in `.`
        """
        self.config.state[SN.output_step.value] = "."
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.name_step.value])
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # Unittest/config/everyvoice-shared-data.yaml
        # Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]), Path("../../Common-Voice")
        )

    def test_wav_file_directory_under_wavs_directory(self):
        """
        output directory is `.`
        wav files directory located in `wavs/`
        """
        self.config.state[SN.output_step.value] = "."
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        wavs_dir = "wavs/Common-Voice"
        self.config.state["dataset_0"][SN.wavs_dir_step.value] = wavs_dir
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.name_step.value])
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # Unittest/config/everyvoice-shared-data.yaml
        # wavs/Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]), Path("../..") / wavs_dir
        )

    def test_output_not_local_and_wav_file_directory_local(self):
        """
        output directory is NOT `.`
        wav files directory located in `.`
        """
        self.config.state[SN.output_step.value] = "John/Smith"
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.output_step.value])
                    / self.config.state[SN.name_step.value]
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # John/Smith/Unittest/config/everyvoice-shared-data.yaml
        # Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]), Path("../../../../Common-Voice")
        )

    def test_output_not_local_and_wav_file_directory_under_hierarchy(self):
        """
        output directory is NOT `.`
        wav files directory located in `wavs/`
        """
        self.config.state[SN.output_step.value] = "John/Smith"
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        wavs_dir = "wavs/Common-Voice"
        self.config.state["dataset_0"][SN.wavs_dir_step.value] = wavs_dir
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.output_step.value])
                    / self.config.state[SN.name_step.value]
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # John/Smith/Unittest/config/everyvoice-shared-data.yaml
        # wavs/Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]),
            Path("../../../..") / wavs_dir,
        )

    def test_absolute_wav_file_directory_and_local_experiment(self):
        """
        output directory is `.`
        wav files directory located in `/ABSOLUTE/wavs/`
        """
        self.config.state[SN.output_step.value] = "."
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                wavs_dir = tmpdir / "wavs/Common-Voice"
                self.config.state["dataset_0"][SN.wavs_dir_step.value] = wavs_dir
                self.config.state["dataset_0"][SN.text_processing_step] = (0,)
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.name_step.value])
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # Unittest/config/everyvoice-shared-data.yaml
        # /tmpdir/wavs/Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]),
            wavs_dir,
        )

    def test_absolute_wav_file_directory_and_nested_experiment(self):
        """
        output directory is NOT `.`
        wav files directory located in `/ABSOLUTE/wavs/`
        """
        self.config.state[SN.output_step.value] = "John/Smith"
        self.config.state[SN.name_step.value] = "Unittest"
        self.config.state.update(CONTACT_INFO_STATE)
        with capture_stdout():
            with tempfile.TemporaryDirectory() as tmpdir:
                os.chdir(tmpdir)
                tmpdir = Path(tmpdir).absolute()
                wavs_dir = tmpdir / "wavs/Common-Voice"
                self.config.state["dataset_0"][SN.wavs_dir_step.value] = wavs_dir
                self.config.state["dataset_0"][SN.text_processing_step] = tuple()
                self.config.effect()
                data_file = (
                    Path(self.config.state[SN.output_step.value])
                    / self.config.state[SN.name_step.value]
                    / "config/everyvoice-shared-data.yaml"
                )
                with data_file.open() as fin:
                    config = yaml.load(fin, Loader=yaml.FullLoader)
        # John/Smith/Unittest/config/everyvoice-shared-data.yaml
        # /tmpdir/wavs/Common-Voice/
        self.assertEqual(
            Path(config["source_data"][0]["data_dir"]),
            wavs_dir,
        )


class TestEnumDict(TestCase):
    """Test the EnumDict class"""

    def test_enum_dict(self):
        """Enum values need to behave the same with or without .value"""
        d = EnumDict()
        d[SN.audio_config_step] = "foo"
        self.assertEqual(d[SN.audio_config_step.value], "foo")
        self.assertEqual(d.get(SN.audio_config_step.value), "foo")

        d[SN.wavs_dir_step.value] = "bar"
        self.assertEqual(d[SN.wavs_dir_step], "bar")
        self.assertEqual(d.get(SN.wavs_dir_step), "bar")

        self.assertEqual(d.get(SN.filelist_format_step, None), None)
        self.assertEqual(d.get(SN.filelist_format_step.value, None), None)

        d.update({SN.contact_email_step: "a@b.com"})
        self.assertEqual(d[SN.contact_email_step.value], "a@b.com")

        self.assertEqual(
            d,
            {
                SN.audio_config_step.value: "foo",
                SN.wavs_dir_step.value: "bar",
                SN.contact_email_step.value: "a@b.com",
            },
        )



import random
from typing import Tuple

import numpy as np
import torch
import torchaudio.transforms as T
from torch.nn.utils.rnn import pad_sequence

from everyvoice.config.preprocessing_config import AudioSpecTypeEnum
from everyvoice.utils import _flatten


def expand(values, durations):
    out = []
    for value, d in zip(values, durations):
        out += [value] * max(0, int(d))
    if isinstance(values, list):
        return np.array(out)
    elif isinstance(values, torch.Tensor):
        return torch.stack(out)
    elif isinstance(values, np.ndarray):
        return np.array(out)


def collate_fn(data):
    # list-of-dict -> dict-of-lists
    # (see https://stackoverflow.com/a/33046935)
    data = [_flatten(x) for x in data]
    data = {k: [dic[k] for dic in data] for k in data[0]}
    for key in data:
        if isinstance(data[key][0], np.ndarray):
            data[key] = [torch.tensor(x) for x in data[key]]
        if torch.is_tensor(data[key][0]):
            data[key] = pad_sequence(data[key], batch_first=True, padding_value=0)
        if isinstance(data[key][0], int):
            data[key] = torch.IntTensor(data[key])
    return data


def dynamic_range_compression_torch(x, C=1, clip_val=1e-5):
    return torch.log(torch.clamp(x, min=clip_val) * C)


def dynamic_range_decompression_torch(x, C=1):
    return torch.exp(x) / C


def get_spectral_transform(
    spec_type,
    n_fft,
    win_length,
    hop_length,
    sample_rate=None,
    n_mels=None,
    f_min=0,
    f_max=8000,
):
    if spec_type == AudioSpecTypeEnum.mel.value:
        return T.MelSpectrogram(
            sample_rate=sample_rate,
            n_fft=n_fft,
            f_min=f_min,
            f_max=f_max,
            win_length=win_length,
            hop_length=hop_length,
            n_mels=n_mels,
            norm="slaney",
            center=True,
        )
    elif spec_type == AudioSpecTypeEnum.mel_librosa.value:
        from librosa.filters import mel as librosa_mel

        transform = T.Spectrogram(
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
            pad=0,
            window_fn=torch.hann_window,
            power=2.0,
            normalized=False,
            center=True,
            pad_mode="reflect",
            onesided=True,
        )
        mel_basis = librosa_mel(
            sr=sample_rate,
            n_fft=n_fft,
            n_mels=n_mels,
            fmin=f_min,
            fmax=f_max,
        )
        mel_basis = torch.from_numpy(mel_basis).float()

        def mel_transform(x):
            transform.to(x.device)
            spec = transform(x)
            sine_windowed_spec = torch.sqrt(spec + 1e-9)
            mel = torch.matmul(mel_basis.to(x.device), sine_windowed_spec)
            return mel

        return mel_transform
    elif spec_type == AudioSpecTypeEnum.linear.value:
        return T.Spectrogram(
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
        )
    elif spec_type == AudioSpecTypeEnum.raw.value:
        return T.Spectrogram(
            n_fft=n_fft,
            win_length=win_length,
            hop_length=hop_length,
            power=None,
        )
    elif spec_type == "istft":
        return T.InverseSpectrogram(
            n_fft=n_fft, win_length=win_length, hop_length=hop_length
        )
    else:
        return None


def get_segments(
    t: torch.Tensor, segment_size: int, start=None
) -> Tuple[torch.Tensor, int]:
    """Randomly select a segment from a tensor, if the segment is too short, pad it with zeros

    Args:
        t (torch.Tensor): A tensor (time as second dimension)
        segment_size (int): segment size (should be in frames if spectrogram input or samples if waveform input)
        start (_type_, optional): start at specific input, otherwise random. Defaults to None.

    Returns:
        Tuple[torch.Tensor, int]: the segment plus the start index of the segment
    """
    t_len = t.size(1)
    if t_len >= segment_size:
        max_start = t_len - segment_size - 1
        if start is not None:
            assert (
                start <= max_start
            ), f"Segment start was set to be {start} but max is {max_start}"
        else:
            start = random.randint(0, max_start)
        t = t[:, start : start + segment_size]
    else:
        start = 0
        t = torch.nn.functional.pad(t, (0, segment_size - t_len), "constant")
    return t, start


def get_device_from_accelerator(accelerator: str) -> torch.device:
    """Given an accelerator name ("auto", "cpu", "gpu", "mps"), return it's torch.device equivalent."""
    match accelerator:
        case "auto":
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        case "gpu":
            device = torch.device("cuda:0")
        case "cpu" | "mps":
            device = torch.device(accelerator)
        case _:
            device = torch.device("cpu")

    return device



from pathlib import Path
from typing import Optional, Union

from pydantic import Field, FilePath, ValidationInfo, model_validator

from everyvoice.config.shared_types import (
    BaseModelWithContact,
    BaseTrainingConfig,
    ContactInformation,
    init_context,
)
from everyvoice.config.utils import PossiblyRelativePath, load_partials
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.utils import load_config_from_json_or_yaml_path


# The contact information only needs to be registered on the main config
class AlignerConfigNoContact(AlignerConfig):
    contact: Optional[ContactInformation] = None


class VocoderConfigNoContact(VocoderConfig):
    contact: Optional[ContactInformation] = None


class FeaturePredictionConfigNoContact(FeaturePredictionConfig):
    contact: Optional[ContactInformation] = None


class E2ETrainingConfig(BaseTrainingConfig):
    feature_prediction_checkpoint: Union[None, PossiblyRelativePath] = None
    vocoder_checkpoint: Union[None, PossiblyRelativePath] = None


class EveryVoiceConfig(BaseModelWithContact):
    aligner: AlignerConfig | AlignerConfigNoContact = Field(
        default_factory=AlignerConfigNoContact
    )
    path_to_aligner_config_file: Optional[FilePath] = None

    feature_prediction: FeaturePredictionConfig | FeaturePredictionConfigNoContact = (
        Field(default_factory=FeaturePredictionConfigNoContact)
    )
    path_to_feature_prediction_config_file: Optional[FilePath] = None

    vocoder: VocoderConfig | VocoderConfigNoContact = Field(
        default_factory=VocoderConfigNoContact
    )
    path_to_vocoder_config_file: Optional[FilePath] = None

    training: E2ETrainingConfig = Field(default_factory=E2ETrainingConfig)
    path_to_training_config_file: Optional[FilePath] = None

    @model_validator(mode="before")  # type: ignore
    def load_partials(self, info: ValidationInfo):
        config_path = (
            info.context.get("config_path", None) if info.context is not None else None
        )
        return load_partials(
            self,  # type: ignore
            ("aligner", "feature_prediction", "vocoder", "training"),
            config_path=config_path,
        )

    @staticmethod
    def load_config_from_path(
        path: Path,
    ) -> "EveryVoiceConfig":
        """Load a config from a path"""
        config = load_config_from_json_or_yaml_path(path)
        with init_context({"config_path": path}):
            config = EveryVoiceConfig(**config)
        return config



from everyvoice.wizard import StepNames, Tour
from everyvoice.wizard.basic import (
    ContactEmailStep,
    ContactNameStep,
    MoreDatasetsStep,
    NameStep,
    OutputPathStep,
)
from everyvoice.wizard.dataset import get_dataset_steps


def get_main_wizard_tour():
    """Get the main wizard tour"""
    return Tour(
        name="Basic Tour",
        steps=[
            NameStep(name=StepNames.name_step),
            ContactNameStep(name=StepNames.contact_name_step),
            ContactEmailStep(name=StepNames.contact_email_step),
            OutputPathStep(name=StepNames.output_step),
        ]
        + get_dataset_steps()
        + [MoreDatasetsStep(name=StepNames.more_datasets_step)],
    )


if __name__ == "__main__":
    get_main_wizard_tour().run()



import json

from everyvoice.base_cli.interfaces import train_base_command_interface
from merge_args import merge_args


@merge_args(train_base_command_interface)
def train(**kwargs):
    from everyvoice.base_cli.helpers import load_config_base_command, train_base_command
    from everyvoice.text.lookups import lookuptables_from_config

    from ..config import FastSpeech2Config
    from ..dataset import FastSpeech2DataModule
    from ..model import FastSpeech2
    from ..type_definitions_heavy import Stats

    config_args = kwargs["config_args"]
    config_file = kwargs["config_file"]
    config = load_config_base_command(FastSpeech2Config, config_args, config_file)
    lang2id, speaker2id = lookuptables_from_config(config)

    # TODO: What about when we are fine-tuning? Do the bins in the Variance Adaptor not change? https://github.com/roedoejet/FastSpeech2_lightning/issues/28
    with open(config.preprocessing.save_dir / "stats.json") as f:
        stats: Stats = Stats(**json.load(f))

    model_kwargs = {"lang2id": lang2id, "speaker2id": speaker2id, "stats": stats}

    train_base_command(
        model_config=FastSpeech2Config,
        model=FastSpeech2,
        data_module=FastSpeech2DataModule,
        monitor="validation/total_loss",
        gradient_clip_val=1.0,
        model_kwargs=model_kwargs,
        **kwargs,
    )



from itertools import chain
from typing import Dict, Iterable, Sequence, Tuple, Union

from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig

LookupTable = dict[str, int]


def lookuptables_from_config(
    config: Union[EveryVoiceConfig, FeaturePredictionConfig]
) -> Tuple[LookupTable, LookupTable]:
    """ """
    train_dataset = config.training.filelist_loader(config.training.training_filelist)
    val_dataset = config.training.filelist_loader(config.training.validation_filelist)

    return lookuptables_from_data((train_dataset, val_dataset))


def lookuptables_from_data(
    data: Iterable[Sequence[Dict[str, str]]]
) -> Tuple[LookupTable, LookupTable]:
    """ """
    languages = set(d["language"] for d in chain(*data) if "language" in d)
    lang2id = {language: i for i, language in enumerate(sorted(languages))}

    speakers = set(d["speaker"] for d in chain(*data) if "speaker" in d)
    speaker2id = {speaker: i for i, speaker in enumerate(sorted(speakers))}

    return lang2id, speaker2id


def build_lookup(items: Sequence[Dict[str, str]], key: str) -> Dict[str, int]:
    """
    Create a lookup table from a list of entries and a key into those entries.
    """
    # Using a dictionary instead of a set to preserve the order.
    uniq_items = {item[key]: 1 for item in items}
    return {item: i for i, item in enumerate(uniq_items.keys())}


class LookupTables:
    def __init__(self, config: Union[EveryVoiceConfig, FeaturePredictionConfig]):
        self.config = config
        self.lang2id, self.speaker2id = lookuptables_from_config(self.config)



# [Valid version numbers](https://packaging.python.org/en/latest/discussions/versioning/#versioning)
# [PEP 440 – Version Identification and Dependency Specification](https://peps.python.org/pep-0440/)
# [Specifying Your Project’s Version](https://setuptools.pypa.io/en/latest/userguide/distribution.html)
# [N!]N(.N)*[{a|b|rc}N][.postN][.devN]
VERSION = "0.1.0a2"



from typing import Dict

from pydantic import BaseModel, ConfigDict, Field, model_validator

from everyvoice.config.shared_types import ConfigModel
from everyvoice.config.utils import PossiblySerializedCallable
from everyvoice.text.utils import normalize_text_helper
from everyvoice.utils import collapse_whitespace


class Punctuation(BaseModel):
    exclamations: list[str] = Field(
        ["!", "¡"],  # TODO: consider how to handle utt final punctuation like ! ? and .
        description="Exclamation punctuation symbols used in your datasets. Replaces these symbols with <EXCL> internally.",
    )
    question_symbols: list[str] = Field(
        ["?", "¿"],
        description="Question/interrogative punctuation symbols used in your datasets. Replaces these symbols with <QINT> internally.",
    )
    quotemarks: list[str] = Field(
        ['"', "'", "“", "”", "«", "»"],
        description="Quotemark punctuation symbols used in your datasets. Replaces these symbols with <QUOTE> internally.",
    )
    big_breaks: list[str] = Field(
        [".", ":", ";"],
        description="Punctuation symbols indicating a 'big break' used in your datasets. Replaces these symbols with <BB> internally.",
    )
    small_breaks: list[str] = Field(
        [",", "-", "—"],
        description="Punctuation symbols indicating a 'small break' used in your datasets. Replaces these symbols with <SB> internally.",
    )
    ellipsis: list[str] = Field(
        ["…"],
        description="Punctuation symbols indicating an ellipsis used in your datasets. Replaces these symbols with <EPS> internally.",
    )

    @property
    def all(self) -> set[str]:
        """Return a set of all punctuations."""
        return (
            set(self.exclamations)
            | set(self.question_symbols)
            | set(self.quotemarks)
            | set(self.big_breaks)
            | set(self.small_breaks)
            | set(self.ellipsis)
        )


class Symbols(BaseModel):
    silence: list[str] = Field(
        ["<SIL>"], description="The symbol(s) used to indicate silence."
    )
    punctuation: Punctuation = Field(
        default_factory=Punctuation,
        description="EveryVoice will combine punctuation and normalize it into a set of five permissible types of punctuation to help tractable training.",
    )
    model_config = ConfigDict(extra="allow")

    @property
    def all_except_punctuation(self) -> set[str]:
        """Returns the set containing all characters."""
        return set(w for _, v in self if not isinstance(v, Punctuation) for w in v)

    @model_validator(mode="after")
    def member_must_be_list_of_strings(self) -> "Symbols":
        """Except for `punctuation` & `pad`, all user defined member variables
        have to be a list of strings.
        """
        for k, v in self:
            if isinstance(v, Punctuation):
                continue
            if k == "pad":
                continue
            if not isinstance(v, list) or not all(isinstance(e, str) for e in v):
                raise ValueError(f"{k} must be a list")

        return self


class TextConfig(ConfigModel):
    symbols: Symbols = Field(default_factory=Symbols)
    to_replace: Dict[str, str] = {}  # Happens before cleaners
    cleaners: list[PossiblySerializedCallable] = [
        collapse_whitespace,
    ]

    @model_validator(mode="after")
    def clean_symbols(self) -> "TextConfig":
        """We should apply all cleaners to the symbols

        Returns:
            TextConfig: a text config with cleaned symbols
        """
        for k, v in self.symbols:
            if k not in ["punctuation", "silence"]:
                setattr(
                    self.symbols,
                    k,
                    [
                        normalize_text_helper(x, self.to_replace, self.cleaners)
                        for x in v
                    ],
                )
        return self



# /*
# *  Arpabet-to-IPA - converting Arpabet to IPA.
# *
# * @author		Waldeilson Eder dos Santos. Adapted by Aidan Pine.
# * @copyright 	Copyright (c) 2015 Waldeilson Eder dos Santos
# * @copyright 	Copyright (c) 2024 National Research Council Canada
# * @link			https://github.com/wwesantos/arpabet-to-ipa
# * @package     	arpabet-to-ipa
# *
# * The MIT License (MIT)
# *
# * Permission is hereby granted, free of charge, to any person obtaining a copy
# * of this software and associated documentation files (the "Software"), to deal
# * in the Software without restriction, including without limitation the rights
# * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# * copies of the Software, and to permit persons to whom the Software is
# * furnished to do so, subject to the following conditions:
# *
# * The above copyright notice and this permission notice shall be included in all
# * copies or substantial portions of the Software.
# *
# * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# * SOFTWARE.
# */

""" A Simple Arpabet to IPA conversion
    adapted by Aidan Pine from https://github.com/wwesantos/arpabet-to-ipa/blob/master/src/App.php

    Except changed so that 1 adds a pre-vocalic primary stress symbol and 2 adds a secondary stress symbol.
    These flout the IPA standard of placing stress symbols before syllables, but this is more computationally
    tractable (in the absence of reliable syllabification algorithms for all datasets that include Arpabet).

"""
from g2p.mappings import Mapping, Rule
from g2p.transducer import Transducer

ARPABET_LOOKUP = {
    "AO": "ɔ",
    "AO0": "ɔ",
    "AO1": "ˈɔ",
    "AO2": "ˌɔ",
    "AA": "ɑ",
    "AA0": "ɑ",
    "AA1": "ˈɑ",
    "AA2": "ˌɑ",
    "IY": "i",
    "IY0": "i",
    "IY1": "ˈi",
    "IY2": "ˌi",
    "UW": "u",
    "UW0": "u",
    "UW1": "ˈu",
    "UW2": "ˌu",
    "EH": "e",  # modern versions use 'e' instead of 'ɛ'
    "EH0": "e",  # ɛ
    "EH1": "ˈe",  # ɛ
    "EH2": "ˌe",  # ɛ
    "IH": "ɪ",
    "IH0": "ɪ",
    "IH1": "ˈɪ",
    "IH2": "ˌɪ",
    "UH": "ʊ",
    "UH0": "ʊ",
    "UH1": "ˈʊ",
    "UH2": "ˌʊ",
    "AH": "ʌ",
    "AH0": "ə",
    "AH1": "ˈʌ",
    "AH2": "ˌʌ",
    "AE": "æ",
    "AE0": "æ",
    "AE1": "ˈæ",
    "AE2": "ˌæ",
    "AX": "ə",
    "AX0": "ə",
    "AX1": "ˈə",
    "AX2": "ˌə",
    # /*
    # Vowels - Diphthongs
    # Arpabet	IPA	Word Examples
    # EY		eɪ	say (S EY1); eight (EY1 T)
    # AY		aɪ	my (M AY1); why (W AY1); ride (R AY1 D)
    # OW		oʊ	show (SH OW1); coat (K OW1 T)
    # AW		aʊ	how (HH AW1); now (N AW1)
    # OY		ɔɪ	boy (B OY1); toy (T OY1)
    # */
    "EY": "eɪ",
    "EY0": "eɪ",
    "EY1": "ˈeɪ",
    "EY2": "ˌeɪ",
    "AY": "aɪ",
    "AY0": "aɪ",
    "AY1": "ˈaɪ",
    "AY2": "ˌaɪ",
    "OW": "oʊ",
    "OW0": "oʊ",
    "OW1": "ˈoʊ",
    "OW2": "ˌoʊ",
    "AW": "aʊ",
    "AW0": "aʊ",
    "AW1": "ˈaʊ",
    "AW2": "ˌaʊ",
    "OY": "ɔɪ",
    "OY0": "ɔɪ",
    "OY1": "ˈɔɪ",
    "OY2": "ˌɔɪ",
    # /*
    # Consonants - Stops
    # Arpabet	IPA	Word Examples
    # P		p	pay (P EY1)
    # B		b	buy (B AY1)
    # T		t	take (T EY1 K)
    # D		d	day (D EY1)
    # K		k	key (K IY1)
    # G		ɡ	go (G OW1)
    # */
    "P": "p",
    "B": "b",
    "T": "t",
    "D": "d",
    "K": "k",
    "G": "g",
    # /*
    # Consonants - Affricates
    # Arpabet	IPA	Word Examples
    # CH		tʃ	chair (CH EH1 R)
    # JH		dʒ	just (JH AH1 S T); gym (JH IH1 M)
    # */
    # 	'CH' : 'tʃ',
    # 	'JH' : 'dʒ',
    # /*
    # Consonants - Fricatives
    # Arpabet	IPA	Word Examples
    # F		f	for (F AO1 R)
    # V		v	very (V EH1 R IY0)
    # TH		θ	thanks (TH AE1 NG K S); Thursday (TH ER1 Z D EY2)
    # DH		ð	that (DH AE1 T); the (DH AH0); them (DH EH1 M)
    # S		s	say (S EY1)
    # Z		z	zoo (Z UW1)
    # SH		ʃ	show (SH OW1)
    # ZH		ʒ	measure (M EH1 ZH ER0); pleasure (P L EH1 ZH ER)
    # HH		h	house (HH AW1 S)
    # */
    "F": "f",
    "V": "v",
    "TH": "θ",
    "DH": "ð",
    "S": "s",
    "Z": "z",
    "SH": "ʃ",
    "ZH": "ʒ",
    "HH": "h",
    # /*
    # Consonants - Nasals
    # Arpabet	IPA	Word Examples
    # M		m	man (M AE1 N)
    # N		n	no (N OW1)
    # NG		ŋ	sing (S IH1 NG)
    # */
    "M": "m",
    "N": "n",
    "NG": "ŋ",
    # /*
    #  Consonants - Liquids
    # Arpabet	IPA		Word Examples
    # L		ɫ OR l	late (L EY1 T)
    # R		r OR ɹ	run (R AH1 N)
    # */
    "L": "l",
    "R": "r",
    # /*
    #  Vowels - R-colored vowels
    # Arpabet			IPA	Word Examples
    # ER				ɝ	her (HH ER0); bird (B ER1 D); hurt (HH ER1 T), nurse (N ER1 S)
    # AXR				ɚ	father (F AA1 DH ER); coward (K AW1 ER D)
    # The following R-colored vowels are contemplated above
    # EH R			ɛr	air (EH1 R); where (W EH1 R); hair (HH EH1 R)
    # UH R			ʊr	cure (K Y UH1 R); bureau (B Y UH1 R OW0), detour (D IH0 T UH1 R)
    # AO R			ɔr	more (M AO1 R); bored (B AO1 R D); chord (K AO1 R D)
    # AA R			ɑr	large (L AA1 R JH); hard (HH AA1 R D)
    # IH R or IY R	ɪr	ear (IY1 R); near (N IH1 R)
    # AW R			aʊr	This seems to be a rarely used r-controlled vowel. In some dialects flower (F L AW1 R; in other dialects F L AW1 ER0)
    # */
    "ER": "ɜr",
    "ER0": "ɜr",
    "ER1": "ˈɜr",
    "ER2": "ˌɜr",
    "AXR": "ər",
    "AXR0": "ər",
    "AXR1": "ˈər",
    "AXR2": "ˌər",
    # /*
    # Consonants - Semivowels
    # Arpabet	IPA	Word Examples
    # Y		j	yes (Y EH1 S)
    # W		w	way (W EY1)
    # */
    "W": "w",
    "Y": "j",
}

mapping = Mapping(
    rules=[Rule(rule_input=k, rule_output=v) for k, v in ARPABET_LOOKUP.items()],
    in_lang="arpabet",
    out_lang="ipa",
    rule_ordering="apply-longest-first",
    out_delimiter="",
)
ARPABET_TO_IPA_TRANSDUCER = Transducer(mapping)



import json
import os
import subprocess
import tempfile
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

import jsonschema
import yaml
from pydantic import ValidationError
from pytorch_lightning import Trainer
from typer.testing import CliRunner
from yaml import CLoader as Loader

# required for `./run_tests.py cli` to work, otherwise test_inspect_checkpoint
# fails with an Intel MKL FATAL ERROR saying it cannot load libtorch_cpu.so
import everyvoice.tests.test_model  # noqa
from everyvoice import __file__ as EV_FILE
from everyvoice.base_cli.helpers import save_configuration_to_log_dir
from everyvoice.cli import SCHEMAS_TO_OUTPUT, app
from everyvoice.config.shared_types import ContactInformation
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.config import (
    FastSpeech2Config,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
    FastSpeech2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.type_definitions_heavy import (
    Stats,
    StatsInfo,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import HiFiGANConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import HiFiGAN
from everyvoice.tests.stubs import capture_logs, mute_logger
from everyvoice.wizard import (
    SPEC_TO_WAV_CONFIG_FILENAME_PREFIX,
    TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX,
)

EV_DIR = Path(EV_FILE).parent


class CLITest(TestCase):
    data_dir = Path(__file__).parent / "data"

    def setUp(self) -> None:
        super().setUp()
        self.runner = CliRunner()
        self.config_dir = Path(__file__).parent / "data" / "relative" / "config"
        self.commands = [
            "new-project",
            "train",
            "synthesize",
            "preprocess",
            "inspect-checkpoint",
        ]

    def wip_test_synthesize(self):
        # TODO: Here's a stub for getting synthesis unit tests working
        #       I believe we'll need to also pass a stats object to the created spec_model
        # TODO: add a test for making sure that `preprocessing` and `logs_and_checkpoints` folders don't get created.
        os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = (
            "1"  # Fallback for running tests on Mac
        )
        import random

        import torch

        # Set a manual seed, because some seeds cause the model
        # to fail to generate a proper wav file. This seed was taken
        # from running torch.seed() on a working run. Note: this is a bit
        # brittle, but this test is just to test that the synthesize command
        # works given two functional checkpoints. Further tests into the effects
        # of seeds should be looked into.
        torch.use_deterministic_algorithms(True)
        torch.manual_seed(10719787423044995460)
        random.seed(10719787423044995460)
        vocoder = HiFiGAN(
            HiFiGANConfig.load_config_from_path(
                self.config_dir / f"{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}.yaml"
            )
        )
        spec_model = FastSpeech2(
            FastSpeech2Config.load_config_from_path(
                self.config_dir / f"{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml"
            ),
            lang2id={"default": 0},
            speaker2id={"default": 0},
            stats=Stats(
                pitch=StatsInfo(
                    min=150, max=300, std=2.0, mean=0.5, norm_max=1.0, norm_min=0.1
                ),
                energy=StatsInfo(
                    min=0.1, max=10.0, std=2.0, mean=0.5, norm_max=1.0, norm_min=0.1
                ),
            ),
        )
        with tempfile.TemporaryDirectory() as tmpdir_str:
            tmpdir = Path(tmpdir_str)
            vocoder_trainer = Trainer(default_root_dir=tmpdir_str, barebones=True)
            fp_trainer = Trainer(default_root_dir=tmpdir_str, barebones=True)
            vocoder_trainer.strategy.connect(vocoder)
            fp_trainer.strategy.connect(spec_model)
            fp_path = tmpdir / "fp.ckpt"
            fp_trainer.save_checkpoint(fp_path)
            vocoder_path = tmpdir / "vocoder.ckpt"
            vocoder_trainer.save_checkpoint(vocoder_path)
            with open(tmpdir / "utt.list", "w") as f:
                f.write(
                    "\n".join(
                        ["this is a test", "here is another test", "and a foo bar test"]
                    )
                )
            single_text_result = self.runner.invoke(
                app,
                [
                    "synthesize",
                    "from-text",
                    str(fp_path),
                    "--vocoder-path",
                    str(vocoder_path),
                    "-o",
                    str(tmpdir / "single_text"),
                    "--text",
                    "hello world",
                    "-O",
                    "wav",
                ],
            )
            self.assertEqual(single_text_result.exit_code, 0)
            self.assertEqual(
                len(list((tmpdir / "single_text" / "wav").glob("*.wav"))), 1
            )  # assert synthesizes a single file
            filelist_result = self.runner.invoke(
                app,
                [
                    "synthesize",
                    "from-text",
                    str(fp_path),
                    "--vocoder-path",
                    str(vocoder_path),
                    "-o",
                    str(tmpdir / "filelist"),
                    "--filelist",
                    str(tmpdir / "utt.list"),
                    "-O",
                    "wav",
                ],
            )
            self.assertEqual(filelist_result.exit_code, 0)
            self.assertEqual(
                len(list((tmpdir / "filelist" / "wav").glob("*.wav"))), 3
            )  # assert synthesizes three files

    def test_commands_present(self):
        result = self.runner.invoke(app, ["--help"])
        # each command has some help
        for command in self.commands:
            self.assertIn(command, result.stdout)
        # link to docs is present
        self.assertIn("https://docs.everyvoice.ca", result.stdout)

    def test_command_help_messages(self):
        for command in self.commands:
            result = self.runner.invoke(app, [command, "--help"])
            self.assertEqual(result.exit_code, 0)
            result = self.runner.invoke(app, [command, "-h"])
            self.assertEqual(result.exit_code, 0)

    def test_update_schema(self):
        dummy_contact = ContactInformation(
            contact_name="Test Runner", contact_email="info@everyvoice.ca"
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            # Validate that schema generation works correctly.
            _ = self.runner.invoke(app, ["update-schemas", "-o", tmpdir])
            for filename, obj in SCHEMAS_TO_OUTPUT.items():
                with open(Path(tmpdir) / filename, encoding="utf8") as f:
                    schema = json.load(f)
                # serialize the model to json and then validate against the schema
                # Some objects will require a contact key
                try:
                    obj_instance = obj()
                except ValidationError:
                    obj_instance = obj(contact=dummy_contact)
                self.assertIsNone(
                    jsonschema.validate(
                        json.loads(obj_instance.model_dump_json()),
                        schema=schema,
                    )
                )

            # Make sure the generated schemas are identical to those saved in the repo,
            # i.e., that we didn't change the models but forget to update the schemas.
            for filename in SCHEMAS_TO_OUTPUT:
                with open(Path(tmpdir) / filename, encoding="utf8") as f:
                    new_schema = f.read()
                try:
                    with open(EV_DIR / ".schema" / filename, encoding="utf8") as f:
                        saved_schema = f.read()
                except FileNotFoundError:
                    raise AssertionError(
                        f'Schema file {filename} is missing, please run "everyvoice update-schemas".'
                    )
                self.assertEqual(
                    saved_schema,
                    new_schema,
                    'Schemas are out of date, please run "everyvoice update-schemas".',
                )

        # Next, but only if everything above passed, we make sure we can't overwrite
        # existing schemas by accident.
        result = self.runner.invoke(app, ["update-schemas"])
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn("FileExistsError", str(result))

    def test_inspect_checkpoint_help(self):
        result = self.runner.invoke(app, ["inspect-checkpoint", "--help"])
        self.assertIn("inspect-checkpoint [OPTIONS] MODEL_PATH", result.stdout)

    def test_inspect_checkpoint(self):
        result = self.runner.invoke(
            app, ["inspect-checkpoint", str(self.data_dir / "test.ckpt")]
        )
        self.assertIn('global_step": 52256', result.stdout)
        self.assertIn(
            "We couldn't read your file, possibly because the version of EveryVoice that created it is incompatible with your installed version.",
            result.stdout,
        )
        self.assertIn("It appears to have 0.0 M parameters.", result.stdout)
        self.assertIn("Number of Parameters", result.stdout)

    def test_preprocessing_with_wrong_config(self):
        """
        The user should have a friendly message that informs them that they used the wrong config file type.
        """
        with capture_logs() as output:
            result = self.runner.invoke(
                app,
                [
                    "preprocess",
                    str(self.config_dir / "everyvoice-aligner.yaml"),
                ],
            )
            self.assertEqual(result.exit_code, 1)
            self.assertIn(
                "We are expecting a FastSpeech2Config but it looks like you provided a DFAlignerConfig",
                output[0],
            )

    def test_expensive_imports_are_tucked_away(self):
        """Make sure expensive imports are tucked away form the CLI help"""
        result = subprocess.run(
            ["everyvoice", "-h"],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            env=dict(os.environ, PYTHONPROFILEIMPORTTIME="1"),
        )

        msg = '\n\nPlease avoid causing {} being imported from "everyvoice -h".\nIt is a relatively expensive import and slows down shell completion.\nRun "PYTHONPROFILEIMPORTTIME=1 everyvoice -h" and inspect the logs to see why it\'s being imported.'
        self.assertNotIn(b"shared_types", result.stderr, msg.format("shared_types.py"))
        self.assertNotIn(b"pydantic", result.stderr, msg.format("pydantic"))


class TestBaseCLIHelper(TestCase):
    def test_save_configuration_to_log_dir(self):
        with TemporaryDirectory() as tempdir, mute_logger(
            "everyvoice.base_cli.helpers"
        ):
            tempdir = Path(tempdir)
            config = FastSpeech2Config(
                contact=ContactInformation(
                    contact_name="Test Runner", contact_email="info@everyvoice.ca"
                ),
                training={
                    "logger": {
                        "save_dir": tempdir / "log",
                        "name": "unittest",
                    },
                },
            )
            save_configuration_to_log_dir(config)

            log_dir = config.training.logger.save_dir / config.training.logger.name
            log = log_dir / "log"
            self.assertTrue(log.exists())

            hparams = log_dir / "hparams.yaml"
            self.assertTrue(hparams.exists())
            with hparams.open(mode="r", encoding="UTF8") as f:
                config_reloaded = yaml.load(f, Loader=Loader)
                self.assertEqual(
                    config.training.logger.save_dir,
                    Path(config_reloaded["training"]["logger"]["save_dir"]),
                )
                self.assertEqual(
                    config.training.logger.name,
                    config_reloaded["training"]["logger"]["name"],
                )

    def test_complete_path_hack(self):
        from everyvoice.base_cli.interfaces import complete_path

        self.assertEqual(complete_path(), [])
